from sympy import *
import numpy as np
import pandas as pd
import itertools
import json

from exercise import Exercise
from step import Step
from reasoningTree import *

class Corrector:

    def __init__(self, max_considered_answers=500, max_considered_errors=3):
        """
        max_considered_answers (int) = max number of answers/feedbacks generated during asynchrone feedback generation
        max_considered_errors (int) = max number of errors that can be introduced in a reasoning
        """
        self.steps_by_weight = self._getStepsByWeight()
        self.max_considered_answers = max_considered_answers
        self.max_considered_errors = max_considered_errors
        self.default_step = Step(is_equation_step=False, generic_expr_before=[sympify(0)], 
                                 generic_expr_after=[sympify(0)], is_error=False, error_message="Default step", weight=1)

    def _getStepsByWeight(self):
        """
        Load reasoning steps and classify them by weight:

        returns:
            list(list(Step))
        """
        data = pd.read_excel(r"./steps.xlsx")
        names = (pd.DataFrame(data, columns=["Name"])).values.tolist()
        weights = (pd.DataFrame(data, columns=["Weight"])).values.tolist()
        is_errors = (pd.DataFrame(data, columns=["Is error"])).values.tolist()
        is_equations = (pd.DataFrame(data, columns=["Is equation"])).values.tolist()
        messages = (pd.DataFrame(data, columns=["Message"])).values.tolist()
        patterns = (pd.DataFrame(data, columns=["Pattern"])).values.tolist()
        before_after = self._patternsToBeforeAfter(patterns)
        errors_by_weight = {}
        correct_steps = []

        for i in range(len(before_after)):
            generic_expr_before = self._getExprFromStr(before_after[i][0])
            generic_expr_after = self._getExprFromStr(before_after[i][1])
            step = Step(name=names[i][0], is_equation_step=is_equations[i][0], generic_expr_before=generic_expr_before,
                        generic_expr_after=generic_expr_after, is_error=is_errors[i][0], error_message=messages[i][0], weight=weights[i][0])
            if step.is_error:
                if step.weight not in errors_by_weight:
                    errors_by_weight[step.weight] = []
                errors_by_weight[step.weight].append(step)
            else:
                correct_steps.append(step)

        steps_by_weight = [correct_steps]
        errors_by_weight_list = sorted(errors_by_weight.items())
        for error_by_weight in errors_by_weight_list:
            steps_by_weight.append(error_by_weight[1])
        return steps_by_weight
    
    def _patternsToBeforeAfter(self, patterns):
        """
        For each step generic pattern in patterns, return separated before and after expressions:

        patterns (list(string)) = the list of the step generic patterns 
        returns:
            list(list(string, string))
        """
        before_after = []
        for pattern in patterns:
            pattern = pattern[0]
            separator_idex = pattern.find("=>")
            before_str = pattern[0: separator_idex]
            after_str = pattern[separator_idex+3: len(pattern)]

            before_after.append([before_str, after_str])

        return before_after

    def _getExprFromStr(self, str_expr):
        """
        Returns an array containing the sympy expression(s) corresponding to the left and right hand sides if equation
        or simply to str_expr if not:

        str_expr (list(string)) = string(s) to convert
        returns:
            array_like(sympy expression)
        """
        try:
            if "=" in str_expr:
                eqal_index = str_expr.find("=")
                left_str = str_expr[0: eqal_index]
                right_str = str_expr[eqal_index+1: len(str_expr)]
                left_expr = sympify(left_str, evaluate=False)
                right_expr = sympify(right_str, evaluate=False)

                return [left_expr, right_expr]
            
            else:
                expression = [sympify(str_expr, evaluate=False)]
                return expression
            
        except Exception as x:
            raise x
    
    def generateExAnswersFeedbacks(self, exercise):
        """
        Generate the feedbacks for exercise accordingly with corrector's parameters and add those feedback to the exercise 
        (asynchronous generation):

        exercise (Exercise) = exercise for which to generate the feedbacks
        returns:
            None
        """
        if type(exercise) is not Exercise:
            raise TypeError("exercise should be an Exercise instance")

        tree = self._generateExReasoningTree(exercise)
        for answer in tree.sol_to_leaves:
            most_probable_errors = self._getMostProbableErrors(tree, answer)
            feedbacks = []
            for error in most_probable_errors:
                step = error["step"]
                before = error["before"]
                after = error["after"]
                
                feedback = {}
                feedback["generic error"] = self._exprToString(step.generic_expr_before) + " => " + self._exprToString(step.generic_expr_after)
                feedback["specific error"] = self._exprToString(before.ex_expression) + " => " + self._exprToString(after.ex_expression)
                feedback["error message"] = step.error_message

                feedbacks.append(feedback)

            exercise.answers_feedbacks[answer] = feedbacks

    def _generateExReasoningTree(self, exercise):
        """
        Generate the reasoning tree of exercise:

        exercise (Exercise) = exercise of whom to generate the tree
        returns:
            ReasoningTree
        """
        root_is_leaf = self._isLeaf(exercise.ex_expr, exercise)
        root = Node(is_leaf=root_is_leaf, associated_step=self.default_step, solutions=exercise.answers,
                    ex_is_equation=True, ex_expression=exercise.ex_expr)
        max_depth = self._getMaxTreeDepth(exercise)
        tree = ReasoningTree(root=root, max_depth=max_depth)

        prev_considered_answer_nb = -1
        error_nb = 0
        while len(tree.sol_to_leaves)<self.max_considered_answers and len(tree.sol_to_leaves)!=prev_considered_answer_nb and error_nb<self.max_considered_errors:
            prev_considered_answer_nb = len(tree.sol_to_leaves)
            last_error_nodes = tree.last_error_nodes
            tree.last_error_nodes = []

            if len(last_error_nodes) == 0:
                self._extendSubTree(tree, 0, root, exercise)
            else:
                np.random.shuffle(last_error_nodes)
                for node in last_error_nodes:
                    self._extendSubTree(tree, node.depth, node, exercise)
            error_nb += 1

        return tree
    
    def _extendSubTree(self, tree, depth, from_node, exercise):
        """
        Extend the tree, strating from from_node:

        tree (ReasoningTree) = the tree to extend
        depth (int) = the depth at which this extension occurs
        from_node (Node) = the node from which to extend the tree
        exercise (Exercise) = the exercise of whom the tree is
        returns:
            None
        """
        if from_node.is_leaf or len(tree.sol_to_leaves) >= self.max_considered_answers or depth >= tree.max_depth:
            return
        
        for step_category in self.steps_by_weight:
            np.random.shuffle(step_category)

            for step in step_category:
                if step.is_equation_step and not exercise.is_equation:
                    continue

                if step.is_equation_step:
                    resulting_expressions = self._getExprFromEquStep(from_node, step)
                else:
                    resulting_expressions = self._getExprFromNonEquStep(from_node, step, exercise.is_equation)
        
                for i in range(len(resulting_expressions)):
                    if len(tree.sol_to_leaves) >= self.max_considered_answers:
                        return
                
                    expr = resulting_expressions[i]
                    is_leaf = self._isLeaf(expr, exercise)
                    solutions = self._getSolutions(expr, exercise)
                    new_node = Node(is_root=False, is_leaf=is_leaf, depth=depth+1, is_correct=solutions==exercise.answers, associated_step=step,
                                    solutions=solutions, parent=from_node, ex_is_equation=exercise.is_equation, ex_expression=expr)
                    
                    tree_node = tree.find(new_node)
                    if tree_node is not None:   #new node already in tree
                        Node.linkParentToChild(from_node, tree_node, step)
                        if not step.is_error and tree_node.depth > depth+1:
                            tree_node.depth = depth+1
                            self._extendSubTree(tree, depth+1, tree_node, exercise)
                    else:   #new node not in tree
                        from_node.children.append(new_node)
                        tree.appendNode(new_node)

                        if step.is_error:   
                            solutions = self._getSolutions(new_node.ex_expression, exercise)
                            self._createLeafFromSol(solutions, new_node, exercise, tree)
                            tree.last_error_nodes.append(new_node)
                        else:
                            self._extendSubTree(tree, depth+1, new_node, exercise)

    def _getExprFromEquStep(self, from_node, step):
        """
        Computes all the ways to apply the equation step to from_node. Returns a list of lists of both left and right hand sides of 
        the resulting expressions:

        from_node (Node) = the node on which to apply step
        step (Step) = the equation step to apply
        returns:
            list(list(sympy expression, sympy expression))
        """
        generic_expr_before = step.generic_expr_before
        generic_expr_after = step.generic_expr_after

        generic_to_actual = {}
        for symb in Add(generic_expr_before[0], generic_expr_before[1], evaluate=False).free_symbols:
            generic_to_actual[symb] = None
        
        for symb in Add(generic_expr_after[0], generic_expr_after[1], evaluate=False).free_symbols:
            if symb not in generic_to_actual:
                return [], []
            
        left_hand_side = from_node.ex_expression[0]
        right_hand_side = from_node.ex_expression[1]

        mappings = self._getMappingsForEquation(generic_to_actual,
                                                generic_expr_before[0], generic_expr_before[1],
                                                left_hand_side, right_hand_side)
        mappings += self._getMappingsForEquation(generic_to_actual,
                                                 generic_expr_before[1], generic_expr_before[0],
                                                 left_hand_side, right_hand_side)
        mappings = self._removeDuplicateMappings(mappings)

        left_expressions_after = self._getExprFromMappings(mappings, generic_expr_after[0], step.name)
        right_expressions_after = self._getExprFromMappings(mappings, generic_expr_after[1], step.name)
    
        resulting_expressions = []
        for i in range(len(left_expressions_after)):
            resulting_expressions.append([left_expressions_after[i], right_expressions_after[i]])

        return resulting_expressions
    
    def _getExprFromNonEquStep(self, from_node, step, exercise_is_equation):
        """
        Computes all the ways to apply the non-equation step to from_node. Returns a list of lists of both left and right hand sides of 
        the resulting expressions if equation and a list of lists containing the resulting expression if not:

        from_node (Node) = the node on which to apply step
        step (Step) = the non-equation step to apply
        exercise_is_equation (boolean) = whether the exercise is an equation or not
        returns:
            list(list(sympy expression))
        """
        resulting_expressions = []

        if exercise_is_equation:
            if step.name=="incorrect factorization" or step.name=="correct fraction simplification" or step.name=="incorrect fraction numerator simplification" or step.name=="incorrect fraction denominator simplification":
                left_expressions = self._applyFactorizationStep(step, from_node.ex_expression[0])
                right_expressions = self._applyFactorizationStep(step, from_node.ex_expression[1])
            else:
                left_expressions = self._applyNonEquationStep(step, from_node.ex_expression[0])
                right_expressions = self._applyNonEquationStep(step, from_node.ex_expression[1])

            for expr in left_expressions:
                resulting_expressions.append([expr, from_node.ex_expression[1]]) 
            for expr in right_expressions:
                resulting_expressions.append([from_node.ex_expression[0], expr])

        else:
            if step.name=="incorrect factorization" or step.name=="correct fraction simplification" or step.name=="incorrect fraction numerator simplification" or step.name=="incorrect fraction denominator simplification":
                expressions = self._applyFactorizationStep(step, from_node.ex_expression[0])
            else:
                expressions = self._applyNonEquationStep(step, from_node.ex_expression[0])
            for expr in expressions:
                resulting_expressions.append([expr])

        return resulting_expressions
    
    def _applyFactorizationStep(self, step, expression):
        """
        Apply incorrect factorization, correct fraction simplification, incorrect fraction numerator simplification
        or incorrect fraction denominator simplification step on expression:

        step (Step) = the step to apply
        expression (sympy expression) = expression on which to apply the step
        returns:
            list(sympy expression)
        """
        generic_expr_after = step.generic_expr_after[0]
        mappings = self._getMappingsForFactorization(expression, step.name)
        expressions_after = self._getExprFromMappings(mappings, generic_expr_after, step.name)
        
        #maybe step can be applied on sub-expressions
        if len(expression.args) != 0:
            if expression.func is Pow:
                sub_expr_after = self._applyFactorizationStep(step, expression.args[0])
                for expr in sub_expr_after:
                    expressions_after.append(Pow(expr, expression.args[1], evaluate=False))
            else:
                for arg in expression.args:
                    remaining_args = list(expression.args)
                    remaining_args.remove(arg)
                    sub_expr_after = self._applyFactorizationStep(step, arg)
                    for expr in sub_expr_after:
                        expressions_after.append(expression.func(expr, *remaining_args, evaluate=False))
                        
        return expressions_after
    
    def _applyNonEquationStep(self, step, expression):
        """
        Computes all the ways to apply the non-equation step to expression:

        step (Step) = the non-equation step to apply
        expression (sympy expression) = expression on which to apply the step
        returns:
            list(sympy expression)
        """
        generic_expr_before = step.generic_expr_before[0]
        generic_expr_after = step.generic_expr_after[0]

        generic_to_actual = {}
        for symb in generic_expr_before.free_symbols:
            generic_to_actual[symb] = None

        mappings = self._getMappings(generic_to_actual, generic_expr_before, expression)
        mappings = self._removeDuplicateMappings(mappings)
        expressions_after = self._getExprFromMappings(mappings, generic_expr_after, step.name)

        #maybe step can be applied on sub-expressions
        if len(expression.args) != 0:
            if expression.func is Pow:
                sub_expr_after = self._applyNonEquationStep(step, expression.args[0])
                for expr in sub_expr_after:
                    expressions_after.append(Pow(expr, expression.args[1]))
            else:
                for arg in expression.args:
                    remaining_args = list(expression.args)
                    remaining_args.remove(arg)
                    sub_expr_after = self._applyNonEquationStep(step, arg)
                    for expr in sub_expr_after:
                        expressions_after.append(expression.func(expr, *remaining_args))

        return expressions_after

    def _getMappingsForEquation(self, mapping, generic_left_expr, generic_right_expr, left_hand_side, right_hand_side):
        """
        Retrieves all possible mappings between the generic expression parameters and the sub-expressions of the equation
        (returns empty list if the equation does not correspond in any way the the generic expression):

        mapping (dict<Symbol, sympy expression>) = dictionary that will map the generic expression's parameters to sub-expressions of the exercise
        generic_left_expr (sympy expression) = left hand side of the generic expression
        generic_right_expr (sympy expression) = right hand side of the generic expression
        left_hand_side (sympy expression) = left hand side of the exercise expression
        right_hand_side (sympy expression) = right hand side of the exercise expression
        returns:
            array(dict<Symbol, sympy expression>)
        """
        left_mappings = self._getMappings(mapping, generic_left_expr, left_hand_side)
        mappings = []

        for mapping in left_mappings:
            mappings += self._getMappings(mapping, generic_right_expr, right_hand_side)

        return mappings
    
    def _getMappingsForFactorization(self, expression, step_name):
        """
        For incorrect factorization, correct fraction simplification, incorrect fraction numerator simplification
        or incorrect fraction denominator simplification step.
        Retrieves all possible mappings between the generic expression parameters and the sub-expressions of expression
        (returns empty list if the equation does not correspond in any way the the generic expression):

        expression (sympy expression) = expression on which to apply the step
        step_name (string) = name of the step
        returns:
            array(dict<Symbol, sympy expression>)
        """
        mappings = []
        if step_name == "incorrect factorization":
            expression = self._addAndMulTogether(expression)
            if expression.func is Add:
                arrangements = self._getArrangements(list(expression.args), 2)
                for arrangement in arrangements:
                    first_term_args = self._factorizeArgs(arrangement[0], Add)
                    second_term_args = self._factorizeArgs(arrangement[1], Add)
                    mappings = self._getMappingsForFactorizationFrom2Parts(first_term_args, second_term_args)
        
        elif step_name=="correct fraction simplification" or step_name=="incorrect fraction numerator simplification" or step_name=="incorrect fraction denominator simplification":
            if expression.func is Rational:
                expression_args = self._getMulArgs(expression)
                expression = Mul(*expression_args, evaluate=False)
            if expression.func is Mul:
                expression = self._addAndMulTogether(expression)
                num_args = list(expression.args)
                denom_args = []
                for arg in num_args:
                    if arg.func is Pow and arg.args[1] == -1:
                        denom_args.append(arg.args[0])
                        num_args.remove(arg)
                if len(denom_args) == 0:
                    return []
                    
                num_args = self._factorizeArgs(num_args, Mul)
                denom_args = self._factorizeArgs(denom_args, Mul)
                mappings = self._getMappingsForFactorizationFrom2Parts(num_args, denom_args)

        return mappings

    def _getMappingsForFactorizationFrom2Parts(self, first_term_args, second_term_args):
        """
        For incorrect factorization, correct fraction simplification, incorrect fraction numerator simplification
        or incorrect fraction denominator simplification step.
        Retrieves all possible mappings of A*B to first_term_args and of A*C to second_term_args:

        first_term_args (list(sympy expression)) = args of the expression that should be mapped to A*B
        second_term_args (list(sympy expression)) = args of the expression that should be mapped to A*C
        returns:
            array(dict<Symbol, sympy expression>)
        """
        mappings = []

        common_args = []
        args = first_term_args.copy()
        for first_term_arg in args:
            if first_term_arg in second_term_args:
                common_args.append(first_term_arg)
                first_term_args.remove(first_term_arg)
                second_term_args.remove(first_term_arg)

        common_args_arrangements = self._getArrangements(common_args, 2)
        for common_args_arrangement in common_args_arrangements:
            if common_args_arrangement[0] == [sympify(1)]:
                continue
            new_mapping = {}
            new_mapping[sympify('A')] = Mul(*common_args_arrangement[0])
            new_mapping[sympify('B')] = Mul(*common_args_arrangement[1], *first_term_args)
            new_mapping[sympify('C')] = Mul(*common_args_arrangement[1], *second_term_args)
            mappings.append(new_mapping)

        return mappings
    
    def _factorizeArgs(self, args, func):
        """
        Factorize the expression obtained by applying func to args and returns the corresponding factors (or [expr, 1] if not factorizable):

        args (list(sympy expression)) = args of the expression to factorize
        func (sympy expression.func) = function of the expression to factorize
        returns:
            list(sympy expression)
        """
        expr = together(func(*args))
        if expr.func is Mul or expr.func is Integer:
            factorized_args = self._getMulArgs(expr)
        else:
            factorized_args = [expr, sympify(1)]

        return factorized_args
    
    def _getMappings(self, mapping, generic_expr, expr):
        """
        Retrieves all possible mappings between generic_expr parameters and the sub-expressions of expr
        (returns empty list if the equation does not correspond in any way the the generic expression):

        mapping (dict<Symbol, sympy expression>) = dictionary that will map the generic expression's parameters to sub-expressions of the exercise
        generic_expr (sympy expression) = generic expression
        expr (sympy expression) = expressions whose sub-expressions are to be mapped to the generic expression parameters
        returns:
            array(dict<Symbol, sympy expression>)
        """
        if len(generic_expr.args) == 0:
            if generic_expr.func is Symbol: #associate expr to generic_expr which is a symbol
                if mapping[generic_expr] is None or str(mapping[generic_expr]) == str(simplify(expr)):
                    mapping[generic_expr] = simplify(expr)
                    return [mapping]
                else:
                    return []
            else:
                if generic_expr == expr:
                    return [mapping]
                else:
                    return []

        if generic_expr.func is not expr.func:
            return []
                
        generic_args = generic_expr.args
        args = []
        for arg in expr.args:
            args += [arg]

        if len(args) < len(generic_args):
            return []
        
        arrangements = []
        if generic_expr.func is Pow:
            base = args[0]
            power = args[1]
            if generic_args[1] == -1 and power < -1:
                base = Pow(base, -power)
                power = sympify(-1)
            arrangements = [[[base], [power]]]
        else:
            arrangements = self._getArrangements(args, len(generic_args))
        mappings = self._getMappingsFromArrangements(arrangements, mapping, generic_expr)

        return mappings
    
    def _addAndMulTogether(self, expression):
        """
        If expression is addition or multiplication, puts nested addition or multiplication togethern (a+(b+c)=>a+b+c):

        expression (sympy expression) = expressions to rework
        returns:
            sympy expression
        """
        if len(expression.args) == 0:
            return expression
        if expression.func != Mul and expression.func != Add:
            return expression
        
        args = []
        for arg in expression.args:
            if arg.func == expression.func:
                for new_arg in self._addAndMulTogether(arg).args:
                    args.append(new_arg)
            else:
                args.append(arg)
                
        return expression.func(*args, evaluate=False)
    
    def _getMulArgs(self, expression):
        """
        Decompose args of multiplication expression into minimal factors (a**2=>a*a, 6=>2*3):

        expression (sympy expression) = expressions to rework
        returns:
            list(sympy expression)
        """
        if expression.func is Integer:
            return self._convertToPrimeProduct(expression)
        elif expression.func is Rational:
            num = self._convertToPrimeProduct(expression.p)
            args = num
            denom = self._convertToPrimeProduct(expression.q)
            args.append(Pow(Mul(*denom, evaluate=False), -1, evaluate=False))
            return args
        elif expression.func is Pow and expression.args[1].func is Integer and expression.args[1]>0:
            args = []
            for i in range(expression.args[1]):
                args.append(expression.args[0])
            return args
        elif expression.func is not Mul:
            return [expression]
        
        args = []
        for arg in expression.args:
            args += self._getMulArgs(arg)

        return args
    
    def _convertToPrimeProduct(self, integer):
        """
        Decompose an integer into its prime factors:

        integer (int) = integer to decompose
        returns:
            list(int)
        """
        remaining = integer
        factors = []
        if remaining < 0:
            factors.append(sympify(-1))
            remaining = -remaining
        if remaining < 4:
            factors.append(remaining)
            return factors
        
        factor = sympify(2)
        while factor <= remaining**(1/2):
            if remaining%factor == 0:
                remaining /= factor
                factors.append(factor)
            else:
                factor += 1
        
        if remaining > 1:
            factors.append(remaining)

        return factors

    def _getArrangements(self, args, nb_boxes):
        """
        Retrieves all possible ways to distribute the elements in args into nb_boxes boxes with at least one element per box:

        args (array(sympy expression)) = list of elements to distribute
        nb_boxes (int) = number of boxes in which to distributes the elements
        returns:
            array(array(array(sympy expression)))
        """
        if nb_boxes == 1:
            return [[args]]
        
        arrangements = {}
        elem_contents, remaining_args = self._chooseAmongArgs(args, len(args)-(nb_boxes-1))
        for i in range(len(elem_contents)):
            other_boxes_arrangement = self._getArrangements(remaining_args[i], nb_boxes-1)
            for arrangement in other_boxes_arrangement:
                arrangements[str([elem_contents[i]] + arrangement)] = [elem_contents[i]] + arrangement

        return list(arrangements.values())

    def _chooseAmongArgs(self, args, nb_max):
        """
        Computes all possible ways to choose between 1 and nb_max elements from args
        Returns a list of all the possibilities and a list of all the corresponding unchosen elements:

        args (array(sympy expression)) = list of elements from which to choose
        nb_max (int) = max number of elements to choose from args
        returns:
            elem_contents (array(array(sympy expression)))
            remaining_args (array(array(sympy expression)))
        """
        elem_contents = []
        remaining_args = []

        for nb in range(1, nb_max+1):
            combinations = list(itertools.combinations(args, nb))

            for comb in combinations:
                elem_contents.append(list(comb))
                
                remaining = args.copy()
                for elem in comb:
                    remaining.remove(elem)
                remaining_args.append(remaining)
        
        return elem_contents, remaining_args
    
    def _getMappingsFromArrangements(self, arrangements, mapping, generic_expr):
        """
        Retrieves all possible mappings between the generic expression parameters and the sub-expressions of the equation from the arrangements
        containing all possible ways to distribute sub-expressions of the exercise to the generic expression's arguments
        (returns empty list if the equation does not correspond in any way the the generic expression):

        arrangements (array(array(array(sympy expression))))
        mapping (dict<Symbol, sympy expression>) = dictionary that will map the generic expression's parameters to sub-expressions of the exercise
        generic_expr (sympy expression) = generic expression
        returns:
            array(dict<Symbol, sympy expression>)
        """
        mappings = []
        for arrangement in arrangements:
            arrangement_mappings = [mapping.copy()]

            for i in range(len(generic_expr.args)):
                updated_mappings = []

                for mapp in arrangement_mappings:
                    expr_to_map = ""
                    if generic_expr.func is Pow:
                        expr_to_map = arrangement[i][0]
                    else:
                        expr_to_map = generic_expr.func(*arrangement[i], evaluate=False)
                    updated_mappings += self._getMappings(mapp, generic_expr.args[i], expr_to_map)

                arrangement_mappings = updated_mappings

            mappings += arrangement_mappings

        return mappings
    
    def _removeDuplicateMappings(self, mappings):
        """
        Removes all duplicates from mappings:

        mappings (array(dict<Symbol, sympy expression>)) = list of mappings
        returns:
            array(dict<Symbol, sympy expression>)
        """
        final_mappings = []
        for mapping in mappings:
            if mapping not in final_mappings:
                final_mappings.append(mapping)

        return final_mappings
    
    def _getExprFromMappings(self, mappings, generic_expr, step_name):
        """
        Computes expressions resulting from substitution of generic expression's parameters with corresponding mapped values:

        mappings (array(dict<Symbol, sympy expression>)) = list of mappings
        generic_expr (sympy expression) = generic expression
        returns:
            array(sympy expression)
        """
        expressions = []
        for mapping in mappings:
            expr = self._rebuildFromGenericAndMapping(generic_expr, mapping, step_name)
            expressions.append(expr)
        
        return expressions

    def _rebuildFromGenericAndMapping(self, generic_expr, mapping, step_name):
        """
        Computes expression resulting from substitution of generic expression's parameters with corresponding mapped values:

        generic_expr (sympy expression) = generic expression
        mapping (dict<Symbol, sympy expression>) = mapping between the generic expression's parameters and corresponding exercise's sub-expressions
        returns:
            array(sympy expression)
        """
        if generic_expr.func is Symbol:
            return mapping[generic_expr]
        
        if len(generic_expr.args) == 0:
            return generic_expr
        
        args = []
        for arg in generic_expr.args:
            args.append(self._rebuildFromGenericAndMapping(arg, mapping, step_name))
        
        eval = True
        if step_name == "incorrect factorization" and generic_expr.func is Mul and str(generic_expr) != "A*C":
            eval = False
        return generic_expr.func(*args, evaluate=eval)
    
    def _isLeaf(self):
        pass

    def _getMostProbableErrors(self, tree, answer):
        """
        Retreives the errors that were introduced in the shortest path to the leaf corresponding to answer and the
        corresponding nodes before and after the error:

        tree (ReasoningTree) = tree to analyze
        answer (string) = answer for which to find the errors
        returns:
            dict('step': Step, 'before': Node, 'after': Node)
        """
        min_weight = -1
        most_probable_errors = []
        for leaf in tree.sol_to_leaves[answer]: 
            most_probable_path = tree.findShortestPathFromRoot(leaf)

            path_errors = []
            path_weight = 0
            for elem in most_probable_path:
                if elem["next step"].is_error:
                    path_weight += elem["next step"].weight
                    path_errors.append({"step": elem["next step"], "before": elem["node"], "after": elem["next node"]})
            
            if min_weight == -1 or path_weight < min_weight:
                min_weight = path_weight
                most_probable_errors = path_errors

        return most_probable_errors
    
    def getFeedbacksFromReasoning(self, ex_json, reasoning):
        """
        Generates feedback based on reasoning (synchronous generation):

        ex_json (json) = json of the exercise that is solved in the reasoning
        reasoning (list(string)) = list of reasoning steps to analyze
        returns:
            dict('generic error': string, 'specific error': string, 'error message': string)
        """
        try:
            ex_dict = json.loads(ex_json)
            exercise = Exercise.fromDict(ex_dict)
            reasoning_expressions = self._getExpressionsFromReasoning(reasoning, exercise)
            reasoning_step_before, reasoning_step_after = self._findErroneousReasoningStep(exercise, reasoning_expressions)

            index = reasoning_expressions.index(reasoning_step_before)
            str_before = reasoning[index]
            str_after = reasoning[index + 1]

            node_before = Node(ex_expression=reasoning_step_before)
            erroneous_solutions = self._getSolutions(reasoning_step_after, exercise)

            #find what erroneous step can lead reasoning_step_before to reasoning_step_after
            for error_category in self.steps_by_weight[1:]:
                for error in error_category:
                    if error.is_equation_step and not exercise.is_equation:
                        continue

                    if error.is_equation_step:
                        resulting_expressions = self._getExprFromEquStep(node_before, error)
                    else:
                        resulting_expressions = self._getExprFromNonEquStep(node_before, error, exercise.is_equation)
                    
                    for expr in resulting_expressions:
                        solutions = self._getSolutions(expr, exercise)
                        if str(solutions) == str(erroneous_solutions):
                            feedback = {}
                            feedback["generic error"] = self._exprToString(error.generic_expr_before) + " => " + self._exprToString(error.generic_expr_after)
                            feedback["specific error"] = str_before + " => " + str_after
                            feedback["error message"] = error.error_message

                            return [feedback]

            feedback = {}
            feedback["generic error"] = ""
            feedback["specific error"] = str_before + " => " + str_after
            feedback["error message"] = "Unable to find the error..."
            return [feedback]
        
        except Exception as x:
            return [str(x)]

    def _getExpressionsFromReasoning(self, reasoning, exercise):
        """
        Converts steps of the reasoning into sympy expressions:

        reasoning (list(string)) = list of reasoning steps to convert
        exercise (Exercise) = exercise that is solved in the reasoning
        returns:
            list(sympy expression)
        """
        try:
            expressions = []
            for step in reasoning:
                self._checkReasoningStep(step)
                expr = self._getExprFromStr(step)
                free_symbols = Add(*expr, evaluate=False).free_symbols
                for sym in free_symbols:
                    if sym not in exercise.params and sym not in exercise.unknowns:
                        raise Exception("Reference to unknown symbol: " + str(sym))
                expressions.append(expr)

            return expressions
        except Exception as x:
            raise x

    def _findErroneousReasoningStep(self, exercise, reasoning):
        """
        Find at which step of the reasoning a mistake has been made:

        reasoning (list(sympy expression)) = list of reasoning steps to analyze
        exercise (Exercise) = exercise that is solved in the reasoning
        returns:
            (sympy expression) = the step at which an error has been made
            (sympy expression) = the step resulting from the error
        """
        prev_expr = exercise.ex_expr
        for expr in reasoning:
            if str(self._getSolutions(expr, exercise)) != str(exercise.answers):
                return prev_expr, expr
            prev_expr = expr

        raise Exception("Unknown error, each step seems fine...")
    
    def _exprToString(self, expression):
        """
        Convert a sympy expression (or 2 expressions if equation) into a string:

        expression (list(sympy expression)) = list of expression(s) to convert
        returns:
            string
        """
        if len(expression) == 1:
            return str(expression[0])
        return str(expression[0]) + " = " + str(expression[1])

class FracSimpCorrector(Corrector):

    def __init__(self, max_considered_answers=500, max_considered_errors=3):
        super().__init__(max_considered_answers=max_considered_answers, max_considered_errors=max_considered_errors)

    def _isLeaf(self, expression, exercise):
        """
        Checks whether expression should be associated to a leaf in the reasoning tree:

        expression (list(sympy expression)) = list containing the expression to check
        exercise => Unused, present for generality purpose
        returns:
            boolean
        """
        expr_str = str(expression[0])
        solution = together(simplify(expression[0]))
        solution_str = str(solution)

        if expr_str == solution_str:
            return True
        return False
    
    def _getSolutions(self, expression, exercise):
        """
        Computes the solution associated to expression:

        expression (list(sympy expression)) = list containing the expression to process
        exercise => Unused, present for generality purpose
        returns:
            list(sympy expression)
        """
        solutions = [together(simplify(simplify(expression[0])))]
        return solutions
    
    def _createLeafFromSol(self, solution, parent, exercise, tree):
        """
        Given the solution, creates a corresponding leaf node and append it to the tree
        (if similar leaf already exists, simply links parent to it):

        solution (list(sympy expression)) = solution to process
        parent (Node) = parent of the new node to be created
        exercise (Exercise) = Exercise for which the tree is being built
        tree (ReasoningTree) = The tree to which to append the new leaf
        returns:
            None
        """
        if len(solution) > 0:
            expression = solution
        else:
            expression = [sympify('0')]
        correct = str(solution) == str(exercise.answers)
        leaf = Node(is_root=False, is_leaf=True, is_correct=correct, solutions=solution, associated_step=self.default_step,
                    parent=parent, ex_is_equation=False, ex_expression=expression)
        
        tree_leaf = tree.find(leaf)
        if tree_leaf is not None:
            Node.linkParentToChild(parent, tree_leaf, self.default_step)
        else:
            parent.children.append(leaf)
            tree.appendNode(leaf)

    def _getMaxTreeDepth(self, exercise):
        """
        Computes the maximum depth to which the tree should be extended for a given exercise:

        exercise (Exercise) = Exercise for which to compute the tree maximum depth
        returns:
            int
        """
        depth = len(exercise.factors) * 2
        return depth
    
    def _checkReasoningStep(self, step):
        """
        Checks whether the given step of reasoning conforms with fraction simplification exercise expected format:

        step (string) = Step to check
        returns:
            None
        """
        if "=" in step:
            raise Exception("This is not an equation, no step should contain the \"=\" symbol!")
        return

class FirstDegEquCorrector(Corrector):

    def __init__(self, max_considered_answers=500, max_considered_errors=3):
        super().__init__(max_considered_answers=max_considered_answers, max_considered_errors=max_considered_errors)

    def _isLeaf(self, expressions, exercise):
        """
        Checks whether expressions should be associated to a leaf in the reasoning tree:

        expressions (list(sympy expression)) = list containing the expressions to check
        exercise (Exercise) = Concerned exercise
        returns:
            boolean
        """
        unknown = exercise.unknowns[0]
        #if the unknown does not appear in the equation, we have an unsolvable expression and no need to expand the branch any further
        if unknown not in Add(expressions[0], expressions[1], evaluate=False).free_symbols:
            return True
        #then check if the equation expression is of the form x = a, a being any token, in which case the equation is solved
        if expressions[0] is unknown:
            other_side = expressions[1]
        elif expressions[1] is unknown:
            other_side = expressions[0]
        else:
            return False
        
        if unknown not in other_side.free_symbols:
            if len(other_side.args) == 0:
                return True
            elif len(exercise.params)>0 and str(other_side)==str(simplify(other_side)):
                return True
        return False
    
    def _getSolutions(self, expressions, exercise):
        """
        Computes the solution associated to expressions:

        expressions (list(sympy expression)) = list containing the expressions to process
        exercise (Exercise) = Concerned exercise
        returns:
            list(sympy expression)
        """
        solutions = solve(Add(expressions[0], -expressions[1]), exercise.unknowns[0])
        return solutions
    
    def _createLeafFromSol(self, solution, parent, exercise, tree):
        """
        Given the solution, creates a corresponding leaf node and append it to the tree
        (if similar leaf already exists, simply links parent to it):

        solution (list(sympy expression)) = solution to process
        parent (Node) = parent of the new node to be created
        exercise (Exercise) = Exercise for which the tree is being built
        tree (ReasoningTree) = The tree to which to append the new leaf
        returns:
            None
        """
        if len(solution) > 0:
            expression = [exercise.unknowns[0], solution[0]]
        else:
            expression = [sympify('0'), sympify('1')]
        correct = str(solution) == str(exercise.answers)
        leaf = Node(is_root=False, is_leaf=True, is_correct=correct, solutions=solution, associated_step=self.default_step,
                    parent=parent, ex_is_equation=True, ex_expression=expression)
        
        tree_leaf = tree.find(leaf)
        if tree_leaf is not None:
            Node.linkParentToChild(parent, tree_leaf, self.default_step)
        else:
            parent.children.append(leaf)
            tree.appendNode(leaf)

    def _getMaxTreeDepth(self, exercise):
        """
        Computes the maximum depth to which the tree should be extended for a given exercise:

        exercise (Exercise) = Exercise for which to compute the tree maximum depth
        returns:
            int
        """
        depth = 1
        depth += self._getExprComplexity(exercise.ex_expr[0])
        depth += self._getExprComplexity(exercise.ex_expr[1])

        return depth

    def _getExprComplexity(self, expr):
        """
        Compute the complexity of expr:

        expr (sympy expression) = Expression to analyse
        returns:
            int
        """
        if len(expr.args) == 0:
            return 1
        
        complexity = 0
        for arg in expr.args:
            complexity += self._getExprComplexity(arg)
        
        return complexity

    def _checkReasoningStep(self, step):
        """
        Checks whether the given step of reasoning conforms with first degree equation exercise expected format:

        step (string) = Step to check
        returns:
            None
        """
        if "=" not in step:
            raise Exception("This is an equation, each step should contain a \"=\" symbol!")
        return