from sympy import *
import numpy as np
import random as rd
import json

from corrector import *
from reasoningTree import *
from exercise import Exercise


rd.seed()

class Generator:

    def __init__(self, corrector, max_int_value=5, nb_param_range=[0, 0], possible_params=[], seed=None):
        """
        corrector (Corrector) = a corrector for this type of exercise to generate the feedback
        max_int_value (int) = max value a generated integer token can be. Does not bound the value a number can take in the final expression but allows to control to which value it can rise to due to multiplications or additions
        nb_param_range (list(int, int)) = tuple of min and max number of parameters in the exercise
        possible_params (list(string)) = list of parameters that can be included in the exercise
        seed (int | float | str | bytes | bytearray) = if not None, manually set the seed of the random generator
        """

        if not isinstance(corrector, Corrector):
            raise TypeError("corrector should be instance of Corrector class")
        
        if type(max_int_value) is not int:
            raise TypeError("max_int_value should be an integer")
        
        try:
            self._checkRangeParameter(nb_param_range, "nb_param_range")
        except Exception as x:
            raise x
            
        if type(possible_params) is not list:
            raise TypeError("possible_params should be a list")
        if len(possible_params) < nb_param_range[1]:
            raise IndexError("possible_params should have a length at least equal to the upper bound in nb_param_range")
        for elem in possible_params:
            if type(elem) is not str or len(elem)<=0:
                raise TypeError("possible_params elements should be non-empty strings")

        
        self.corrector = corrector
        self.max_int_value = max_int_value
        self.nb_param_range = nb_param_range
        self.possible_params = possible_params

        if seed is not None:
            rd.seed(seed)

    def _checkRangeParameter(self, param, param_name):
        """
        Check that the given range parameter is valid. If not, raise an error:

        param (list(int, int)) = parameter value
        param_name (string) = name of the parameter being checked
        """
        if type(param) is not list:
            raise TypeError(param_name + " should be a list")
        if len(param)!=2:
            raise IndexError(param_name + " should have a length of 2")
        for elem in param:
            if type(elem) is not int or elem<0:
                raise TypeError(param_name + " elements should be positive integers")
        if param[1] < param[0]:
            raise ArithmeticError(param_name + " second element should be >= than the first element")

    def _drawSymbols(self):
        '''
        from nb_param_range and possible_params, draw some parameters:

        returns:
            symbol_list (list(sympy.Symbol))
        '''
        nb_params = rd.randint(self.nb_param_range[0], self.nb_param_range[1])
        symbol_list = self._generateSymbols(nb_params)

        return symbol_list

    def _generateSymbols(self, nb_params):
        '''
        generate a list of nb_params symbols:

        nb_params (int)
        returns:
            list(sympy.Symbol)
        '''
        symbol_list = np.array([])
        for i in range(nb_params):
            sym = symbols(self.possible_params[i])
            symbol_list = np.append(symbol_list, sym)

        return symbol_list
    
    def _generateComplexExpression(self, symbol_list=np.array([]), complexity=0, operation=[Add,Mul,Pow],
                                   operation_proba=[0.4,0.4,0.2], pow_range=(2,2), fraction_proba=0, allow_zero=False, allow_1=True):
        '''
        Generate an expression composed of several subexpresion:

        symbol_list (list(sympy.Symbol)) = list of parameters that may appear in the expression
        complexity (int) = number of subexpressions that will compose the final expression
        operation (list(sympy operator)) = list of operations that are allowed in the expression
        operation_proba (list(float)) = list of probabilities of having a the corresponding element in operation list as operation for each subexpression
        pow_range (tuple(int, int)) = tuple of min and max values of power when applying Power operation
        fraction_proba (float) = probability fo each new token to be a simple fraction
        allow_zero (boolean) = sets whether tokens can be of value 0
        allow_1 (boolean) = sets whether tokens can be of value 1
        returns:
            sympy expression
        '''

        if complexity == 0:
            return self._generateRandomToken(symbol_list, fraction_proba=fraction_proba, allow_zero=allow_zero, allow_1=allow_1)
        
        operation_offset = np.array([])
        prev = 0.
        for i in range(len(operation_proba)):
            operation_offset = np.append(operation_offset, prev + operation_proba[i])
            prev = operation_offset[i]

        expr_type = rd.uniform(0, 1)
        for i in range(len(operation_offset)):
            if expr_type <= operation_offset[i]:
                if operation[i] == Pow:
                    expr = self._generateComplexExpression(symbol_list=symbol_list, complexity=complexity-1, operation=operation,
                                                           operation_proba=operation_proba, pow_range=pow_range,
                                                           fraction_proba=fraction_proba, allow_zero=(allow_zero and complexity-1!=0),
                                                           allow_1=(allow_1 and complexity-1!=0))
                    power = rd.randint(pow_range[0], pow_range[1])
                    return Pow(expr, power)
                else:
                    return self._generateRandomExpr(operation[i], symbol_list=symbol_list, complexity=complexity-1,
                                                    operation=operation, operation_proba=operation_proba, pow_range=pow_range,
                                                    fraction_proba=fraction_proba, allow_zero=allow_zero, allow_1=allow_1)

    def _generateRandomExpr(self, fun, symbol_list=np.array([]), complexity=0, operation=[Add,Mul,Pow],
                            operation_proba=[0.4,0.4,0.2], pow_range=(2,2), fraction_proba=0, allow_zero=False, allow_1=True):
        '''
        Generate an expression based on a given operator:

        fun (sympy operator) = operator of the expression
        symbol_list (list(sympy.Symbol)) = list of parameters that may appear in the expression
        complexity (int) = sum of complexities of expression operands, number of subexpressions that will compose the final expression's operands
        operation (list(sympy operator)) = list of operations that are allowed in the subexpressions
        operation_proba (list(float)) = list of probabilities of having a the corresponding element in operation list as operation for each subexpression
        pow_range (tuple(int, int)) = tuple of min and max values of power when applying Power operation
        fraction_proba (float) = probability fo each new token to be a simple fraction
        allow_zero (boolean) = sets whether tokens can be of value 0
        allow_1 (boolean) = sets whether tokens can be of value 1
        returns:
            sympy expression
        '''
        
        first_arg_complexity = rd.randint(0, complexity)
        second_arg_complexity = complexity - first_arg_complexity
        first_arg = self._generateComplexExpression(symbol_list, first_arg_complexity, operation=operation,
                                                    operation_proba=operation_proba, pow_range=pow_range,
                                                    fraction_proba=fraction_proba, allow_zero=allow_zero,
                                                    allow_1=(allow_1 and (first_arg_complexity!=0 or fun==Add)))
        second_arg = self._generateComplexExpression(symbol_list, second_arg_complexity, operation=operation,
                                                     operation_proba=operation_proba, pow_range=pow_range,
                                                     fraction_proba=fraction_proba, allow_zero=allow_zero,
                                                     allow_1=(allow_1 and (second_arg_complexity!=0 or fun==Add)))

        return fun(first_arg, second_arg)

    def _generateRandomToken(self, symbol_list, fraction_proba=0, allow_zero=False, allow_1=True):
        '''
        Generate a token, the simplest expression possible, may be a simple fraction:

        symbol_list (list(sympy.Symbol)) = list of parameters that may appear in the token
        fraction_proba (float) = probability fo new token to be a simple fraction
        allow_zero (boolean) = sets whether token can be of value 0
        allow_1 (boolean) = sets whether token can be of value 1
        returns:
            sympy expression
        '''
        #choose if fraction or not
        rand = rd.uniform(0, 1)
        if rand < fraction_proba:
            num = self._generateRandomToken(symbol_list)
            denom = self._generateRandomToken(symbol_list, allow_1=False)
            token = Mul(num, Pow(denom, -1))

            if num.is_number and denom.is_number and num % denom == 0:
                add_nb = rd.randint(1, abs(denom)-1)
                num += add_nb * rd.choice([-1, 1])

                token = Mul(num, Pow(denom, -1))

            return token

        #choose a parameter or a number    
        index = rd.randint(0, len(symbol_list))
        if index == len(symbol_list):
            if allow_zero:
                min_val = 0
            else:
                if allow_1:
                    min_val = 1
                else:
                    min_val = 2
            token = sympify(rd.randint(min_val, self.max_int_value))

        else:
            token = symbol_list[index]

        #choose whether to make the token negative or not
        negative = rd.randint(0, 1)
        if negative:
            return -token
        else:
            return token
        
    def _ensureAllParamsPresent(self, expressions, param_list):
        '''
        Check that each parameter in param_list appears in expressions. If not, insert the missing parameters into expressions:

        expressions (list(sympy expression)) = list containing the expression (or 2 expressions if equation)
        param_list (list(sympy.Symbol)) = list of parameters that should appear in expressions
        returns:
            sympy expression
        '''
        global_expr = Add(*expressions, evaluate=False)
        used_params = global_expr.free_symbols
        for param in param_list:
            if param not in used_params:
                index = rd.choice(range(len(expressions)))
                expressions[index] = self._insertParam(expressions[index], param)

        return expressions

    def _insertParam(self, expr, param):
        '''
        Insert param somewhere in expr:

        expr (sympy expression) = expression in which to insert the parameter
        param (sympy.Symbol) = parameter to insert
        returns:
            sympy expression
        '''
        poss_places = self._countTokens(expr)
        if poss_places > 0:
            place = rd.randint(1, poss_places)
            expr, _ = self._insertParamAtPlace(expr, place, param)
            return expr
        else:
            return expr + param


    def _countTokens(self, expr):
        '''
        Count tokens in expr:

        expr (sympy expression) = expression of whom to count tokens
        returns:
            int
        '''
        if len(expr.args) == 0:
            return 1
        
        nb = 0
        if expr.func == Pow:
            nb += self._countTokens(expr.args[0])
        else:
            for arg in expr.args:
                nb += self._countTokens(arg)

        return nb
    
    def _insertParamAtPlace(self, expr, place, param):
        '''
        Insert param in expr. Place indicates at which of the possible possible positions the parameter must be inserted:

        expr (sympy expression) = expression in which to insert the parameter
        param (sympy.Symbol) = parameter to insert
        place (int) = position at which to insert the parameter
        returns:
            sympy expression
        '''
        if len(expr.args) == 0:
            place -= 1
            if place == 0:
                expr = Add(expr, param)
                place -= 1
            return expr, place
        
        else:
            args = expr.args
            new_args = [0 for i in range(len(args))]
            if expr.func == Pow:
                new_args[0], place = self._insertParamAtPlace(args[0], place, param)
                new_args[1] = args[1]
            else:
                for i in range(len(args)):
                    new_args[i], place = self._insertParamAtPlace(args[i], place, param)
                    
            return expr.func(*new_args), place



class FracSimpGenerator(Generator):

    def __init__(self, config_file, corrector, config=None, seed=None):
        '''
        config_file (opened file) = file containing the configuration for this generator if config is None
        corrector (FracSimpCorrector) = a FracSimpCorrector instance to generate the feedback
        config (dict) = a dictionary containing the configuration for this generator
        seed (int | float | str | bytes | bytearray) = if not None, manually set the seed of the random generator

        nb_factor_range (list(int, int)) =  tuple of min and max number of factors that can be factorized
        answer_num_complexity_range (list(int, int)) = tuple of min and max complexity of the answer's numerator
        answer_denom_complexity_range (list(int, int)) = tuple of min and max complexity of the answer's denominator
        factor_complexity_range (list(int, int)) = tuple of min and max complexity of each factor by which the fraction can be simplified
        '''
        if not isinstance(corrector, FracSimpCorrector):
            raise TypeError("corrector should be instance of FracSimpCorrector class")

        if config is None:
            config = json.load(config_file)
        
        frac_simp_config = config['frac_simplification']
        
        try:
            self._checkRangeParameter(frac_simp_config['nb_factor_range'], "nb_factor_range")
            self._checkRangeParameter(frac_simp_config['answer_num_complexity_range'], "answer_num_complexity_range")
            self._checkRangeParameter(frac_simp_config['answer_denom_complexity_range'], "answer_denom_complexity_range")
            self._checkRangeParameter(frac_simp_config['factor_complexity_range'], "factor_complexity_range")
            
            super().__init__(corrector, max_int_value=config['max_int_value'], nb_param_range=config['nb_param_range'],
                             possible_params=config['possible_params'], seed=seed)
        except Exception as x:
            raise x
        
        self.nb_factor_range = frac_simp_config['nb_factor_range']
        self.answer_num_complexity_range = frac_simp_config['answer_num_complexity_range']
        self.answer_denom_complexity_range = frac_simp_config['answer_denom_complexity_range']
        self.factor_complexity_range = frac_simp_config['factor_complexity_range']

    def generateEx(self, seed=None):
        '''
        Randomly generate an exercise accordingly with the generator's configuration. Returns the exercise in its json format:

        seed (int | float | str | bytes | bytearray) = if not None, manually set the seed of the random generator
        returns:
            json
        '''

        if seed is not None:
            rd.seed(seed)
        
        nb_factors = rd.randint(self.nb_factor_range[0], self.nb_factor_range[1])
        symbol_list = self._drawSymbols()

        answer = self._generateExAnswer(symbol_list)
        complex_fraction, factors = self._complexifyFraction(symbol_list, answer, nb_factors)

        exercise = Exercise(params=symbol_list, ex_expr=[complex_fraction], factors=factors, answers=[simplify(answer)],
                            prt_message="Simplify the following fraction as much as possible: ")
        self.corrector.generateExAnswersFeedbacks(exercise)
        ex_as_dict = exercise.asDict()
        ex_as_json = json.dumps(ex_as_dict)

        return ex_as_json
    
    def _generateExAnswer(self, symbol_list):
        '''
        Generates a fraction which will be the answer to the exercise:

        symbol_list (list(sympy.Symbol)) = list of parameters that may appear in the answer
        returns:
            sympy expression
        '''

        num_complexity = rd.randint(self.answer_num_complexity_range[0], self.answer_num_complexity_range[1])
        denom_complexity = rd.randint(self.answer_denom_complexity_range[0], self.answer_denom_complexity_range[1])
        
        answer = self._generateComplexAnswer(symbol_list, num_complexity, denom_complexity)
        answer = simplify(simplify(answer))
        answer = simplify(together(answer))
        
        [answer] = self._ensureAllParamsPresent([answer], symbol_list)
        answer = simplify(answer)

        return answer

    def _generateComplexAnswer(self, symbol_list, num_complexity, denom_complexity):
        '''
        Generates a fraction of two complex expressions:

        symbol_list (list(sympy.Symbol)) = list of parameters that may appear in the answer
        num_complexity (int)
        denom_complexity (int)
        returns:
            sympy expression
        '''
        numerator, denominator = 0, 0
        while numerator == 0:
            numerator = self._generateComplexExpression(symbol_list=symbol_list, complexity=num_complexity)
        while denominator == 0:
            denominator = self._generateComplexExpression(symbol_list=symbol_list, complexity=denom_complexity)

        return numerator/denominator
    
    def _complexifyFraction(self, symbol_list, fraction, nb_factors):
        '''
        Complexifies a fracion by multiplying both numerator and denominator with some factors:

        symbol_list (list(sympy.Symbol)) = list of parameters that may appear in the fraction
        fraction (sympy expression) = fraction to complexify
        nb_factors (int) = number of factors with which to complexify the fraction
        returns:
            sympy expression
        '''
        numerator = numer(fraction)
        denominator = denom(fraction)
        factors = []
        
        i = 0
        while i < nb_factors:
            complexity = rd.randint(self.factor_complexity_range[0], self.factor_complexity_range[1])
            factor = self._generateComplexExpression(symbol_list, complexity, allow_1=complexity!=0)

            if factor != 0:
                numerator = Mul(factor, numerator)
                denominator = Mul(factor, denominator)
                factors.append(factor)
            else:
                i -= 1
            i += 1

        numerator = expand(numerator)
        denominator = expand(denominator)
        complex_faction = sympify("("+str(numerator)+")"+"/"+"("+str(denominator)+")", evaluate=False)

        return complex_faction, factors


class FirstDegEquGenerator(Generator):

    def __init__(self, config_file, corrector, config=None, seed=None):
        '''
        config_file (opened file) = file containing the configuration for this generator if config is None
        corrector (FirstDegEquCorrector) = a FracSimpCorrector instance to generate the feedback
        config (dict) = a dictionary containing the configuration for this generator
        seed (int | float | str | bytes | bytearray) = if not None, manually set the seed of the random generator

        possible_unknowns (list(str)) = list of the possible unknowns that can be used
        fraction_token_proba (float) = probability for each considered subexpression to be a fraction with unknown independent denominator
        complex_factor_proba (float) = probability for each a*x+b factor to be expanded into a a*(b*x+c)+d term
        add_unknown_terms_range (list(int, int)) = min and max number of additional unknown dependent terms
        add_unknown_factors_range (list(int, int)) = min and max number of additional unknown dependent factors that may appear inside an unknown dependent term, while the equation is insured to remain of the first order
        unknown_division_proba (float) = probability of unknown dependent factor being at the denominator
        '''
        if not isinstance(corrector, FirstDegEquCorrector):
            raise TypeError("corrector should be instance of FirstDegEquCorrector class")

        if config is None:
            config = json.load(config_file)
            
        equation_config = config['first_degree_equation']
            
        if type(equation_config['possible_unknowns']) is not list:
            raise TypeError("possible_unknowns should be a list")
        if len(equation_config['possible_unknowns']) < 1:
            raise IndexError("possible_unknowns should contain at least one element")
        for elem in equation_config['possible_unknowns']:
            if type(elem) is not str or len(elem)<=0:
                raise TypeError("possible_unknowns elements should be non-empty strings")
            
        if equation_config['fraction_token_proba']>1 or equation_config['fraction_token_proba']<0:
            raise ArithmeticError("fraction_token_proba should be >=0 and <=1")
            
        if equation_config['complex_factor_proba']>1 or equation_config['complex_factor_proba']<0:
            raise ArithmeticError("complex_factor_proba should be >=0 and <=1")
            
        if equation_config['unknown_division_proba']>1 or equation_config['unknown_division_proba']<0:
            raise ArithmeticError("unknown_division_proba should be >=0 and <=1")
        
        try:
            self._checkRangeParameter(equation_config['add_unknown_terms_range'], "add_unknown_terms_range")
            self._checkRangeParameter(equation_config['add_unknown_factors_range'], "add_unknown_factors_range")

            super().__init__(corrector, max_int_value=config['max_int_value'], nb_param_range=config['nb_param_range'],
                             possible_params=config['possible_params'], seed=seed)
        except Exception as x:
            raise x
        
        self.possible_unknowns = equation_config['possible_unknowns']
        self.fraction_token_proba = equation_config['fraction_token_proba']
        self.complex_factor_proba = equation_config['complex_factor_proba']
        self.add_unknown_terms_range = equation_config['add_unknown_terms_range']
        self.add_unknown_factors_range = equation_config['add_unknown_factors_range']
        self.unknown_division_proba = equation_config['unknown_division_proba']

    def generateEx(self, seed=None):
        '''
        Randomly generate an exercise accordingly with the generator's configuration. Returns the exercise in its json format:

        seed (int | float | str | bytes | bytearray) = if not None, manually set the seed of the random generator
        returns:
            json
        '''

        if seed is not None:
            rd.seed(seed)

        param_list = self._drawSymbols()
        unknown, param_list = self._generateUnknown(param_list)
        
        left_hand_side, right_hand_side = self._generateFirstDegEquation(param_list, unknown)
        answer = solve(simplify(left_hand_side - right_hand_side), unknown)                   #left_hand_side - right_hand_side = homogeneous equation

        exercise = Exercise(params=param_list, is_equation=True, ex_expr=[left_hand_side, right_hand_side], unknowns=[unknown],
                            answers=answer, prt_message="Solve the following equation with regard to " + str(unknown) + ": ")
        self.corrector.generateExAnswersFeedbacks(exercise)
        ex_as_dict = exercise.asDict()
        ex_as_json = json.dumps(ex_as_dict)

        return ex_as_json

    def _generateUnknown(self, param_list):
        '''
        Draws the unknown of the equation and its parameters:

        param_list (list(sympy.Symbol)): list of parameters that may appear in the equation
        returns:
            unknown (sympy.Symbol)
        '''

        unknown_index = rd.randint(0, len(self.possible_unknowns)-1)
        unknown = symbols(self.possible_unknowns[unknown_index])

        if unknown in param_list:
            if len(param_list) < len(self.possible_params):
                sym = symbols(self.possible_params[len(param_list)])
                param_list = np.append(param_list, sym)

            param_list = np.setdiff1d(param_list, unknown)
        
        return unknown, param_list

    def _generateFirstDegEquation(self, param_list, unknown):
        '''
        Generates a first degree equation accordingly to the generator's parameters:

        unknown (sympy.Symbol) = unknown of the equation
        param_list (list(sympy.Symbol)) = list of parameters that may appear in the equation
        returns:
            left_hand_side (sympy expression)
            right_hand_side (sympy expression)
        '''

        left_hand_side = Integer(0)
        right_hand_side = Integer(0)
        unknown_term_nb = 1
        unknown_term_nb += rd.randint(self.add_unknown_terms_range[0], self.add_unknown_terms_range[1])

        #Generate unknown dependent terms
        for i in range(unknown_term_nb):
            left_hand_side, right_hand_side = self._addNewUnknownTerm(left_hand_side, right_hand_side, param_list, unknown, unknown_term_nb-i-1)
        
        #If one side is zero (through chance), replace it with a random token for greater diversity
        if left_hand_side == 0:
            left_hand_side = self._generateRandomToken(param_list, fraction_proba=self.fraction_token_proba, allow_zero=True)
        elif right_hand_side == 0:
            right_hand_side = self._generateRandomToken(param_list, fraction_proba=self.fraction_token_proba, allow_zero=True)

        if len(param_list) != 0:
            left_hand_side, right_hand_side = self._ensureAllParamsPresent([left_hand_side, right_hand_side], param_list)

        if self.nb_param_range == [0, 0] and self.fraction_token_proba == 0 and self.add_unknown_terms_range == [0, 0] and self.add_unknown_factors_range == [0, 0]:
            #If requested equation is of the simplest form (a*(b*x+c)+d = e), avoid the answer being a fraction
            left_hand_side, right_hand_side = self._makeAnswerSimple(left_hand_side, right_hand_side, unknown)

        return left_hand_side, right_hand_side
    
    def _addNewUnknownTerm(self, left_hand_side, right_hand_side, param_list, unknown, remaining_terms):
        '''
        Generates a term accordingly to the generator's parameters and adds it to one side of the equation:
        
        left_hand_side (sympy expression) = expression of the left hand side of the equation
        right_hand_side (sympy expression) = expression of the right hand side of the equation
        param_list (list(sympy.Symbol)) = list of parameters that may appear in the equation
        unknown (sympy.Symbol) = unknown of the equation
        remaining_terms (int) = number of terms that are yet to be added after this one
        returns:
            left_hand_side (sympy expression)
            right_hand_side (sympy expression)
        '''
        #The possibility of unkown at the denominator is ignored for now and the corresponding code diabled
        if remaining_terms > 0:
            nb_factors = 1
            nb_factors += rd.randint(self.add_unknown_factors_range[0], self.add_unknown_factors_range[1])
            term = 1
            for i in range(nb_factors):
                term = Mul(term, self._generateUnknownFactor(param_list, unknown), evaluate=term==1)
            
            side = rd.randint(0, 1)
            if side:
                left_hand_side = Add(left_hand_side, term, evaluate=left_hand_side==0)
            else:
                right_hand_side = Add(right_hand_side, term, evaluate=right_hand_side==0)
                
        else:   #Last term should be such that powers of x greater than 1 cancel out
            homogeneous_expr = left_hand_side - right_hand_side
            coeffs = Poly(simplify(numer(homogeneous_expr)), unknown).all_coeffs()
            equ_degree = len(coeffs) - 1
            term = 0
            for i in range(equ_degree-1):
                term += coeffs[i]*Pow(unknown, equ_degree-i)

            term += self._generateUnknownFactor(param_list, unknown)
            
            side = rd.randint(0, 1)
            if side:
                left_hand_side = Add(left_hand_side, -term, evaluate=left_hand_side==0)

            else:
                right_hand_side = Add(right_hand_side, term, evaluate=right_hand_side==0)

        #This code is takes the possibility of having fractions with unknown dependent denominators, still needs to be debugged (and refactorized?)
        """numer_homo_coeff, denom_homo_coeff = self._getHomogeneousCoeff(left_hand_side, right_hand_side, unknown)
        equ_degree = len(numer_homo_coeff) - 1

        if remaining_terms >= equ_degree-1:

            term = self._generateRandomToken(param_list, fraction_proba=self.fraction_token_proba)
            nb_factors = 1
            nb_factors += rd.randint(self.add_unknown_factors_range[0], self.add_unknown_factors_range[1])

            for i in range(nb_factors):
                updated_term = self._addRegularUnknownFactor(term, denom_homo_coeff, param_list, unknown, remaining_terms, equ_degree)
                if updated_term == -1:
                    break
                term = updated_term
            
            side = rd.randint(0, 1)
            if side:
                left_hand_side += term
            else:
                right_hand_side += term

        else:

            term = Integer(1)
            num_denom_order_diff = len(numer_homo_coeff) - len(denom_homo_coeff)
            term = self._numerAddFactors(term, num_denom_order_diff, param_list, unknown)

            missing_homo_denom_factors = self._getFactorsToCommonDenom(term, denom_homo_coeff, unknown)
            term = self._decreaseUnifiedNumerOrder(-num_denom_order_diff, term, missing_homo_denom_factors)

            term = self._complexifyTerm(term, denom_homo_coeff, param_list, unknown)

            term = self._normalizeToCancelGreatestOrder(numer_homo_coeff[0], term, denom_homo_coeff, unknown)
            side = rd.randint(0, 1)
            if side:
                left_hand_side -= term
            else:
                right_hand_side += term"""

        return left_hand_side, right_hand_side
    
    def _generateUnknownFactor(self, param_list, unknown):
        '''
        Generates an unknown dependent factor of the form a*x+b (a may be 1 and b 0) or c*(a*x+b)+d (a may be 1 and d 0)
        accordingly with the generator's parameters:

        param_list (list(sympy.Symbol)) = list of parameters that may appear in the equation
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            sympy expression
        '''
        rand = rd.uniform(0, 1)
        is_complex_factor = rand < self.complex_factor_proba

        a = self._generateRandomToken(param_list, fraction_proba=self.fraction_token_proba)
        b = self._generateRandomToken(param_list, fraction_proba=self.fraction_token_proba, allow_zero= not is_complex_factor)
        factor = a*unknown + b

        if is_complex_factor:
            c = self._generateRandomToken(param_list, fraction_proba=self.fraction_token_proba, allow_1=False)
            d = self._generateRandomToken(param_list, fraction_proba=self.fraction_token_proba, allow_zero=True)
            factor = Add(Mul(c, factor, evaluate=False), d, evaluate= d==0)

        return factor
    
    def _makeAnswerSimple(self, left_hand_side, right_hand_side, unknown):
        '''
        Adjusts the equation so that the answer is not an integer (does not work if there are parameters):

        left_hand_side (sympy expression) = expression of the left hand side of the equation
        right_hand_side (sympy expression) = expression of the right hand side of the equation
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            left_hand_side (sympy expression)
            right_hand_side (sympy expression)
        '''
        
        homogeneous_expr = left_hand_side - right_hand_side
        coeffs = Poly(simplify(numer(homogeneous_expr)), unknown).all_coeffs()
        sign = rd.choice([-1, 1])
        adjust = coeffs[1] % coeffs[0]
        multiplicator = rd.randint(0, self.max_int_value)
        side = rd.randint(0, 1)
        if side:
            left_hand_side -= adjust + multiplicator * coeffs[0] * sign
        else:
            right_hand_side += adjust + multiplicator * coeffs[0] * sign
            
        return left_hand_side, right_hand_side
    

    #These functions are used in case of unknown dependent denominators
    """def _getHomogeneousCoeff(self, left_hand_side, right_hand_side, unknown):
        '''
        Reduces the equation to its homogeneous form and computes the coefficients of both numerator and denominator polynomials:

        left_hand_side (sympy expression) = expression of the left hand side of the equation
        right_hand_side (sympy expression) = expression of the right hand side of the equation
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            numer_homo_coeff (list[token])
            denom_homo_coeff (list[token])
        '''

        homogeneous_equation = (left_hand_side - right_hand_side).together()
        numer_homo_coeff = Poly(numer(homogeneous_equation), unknown).all_coeffs()
        denom_homo_coeff = Poly(denom(homogeneous_equation), unknown).all_coeffs()

        return numer_homo_coeff, denom_homo_coeff
    
    def _addRegularUnknownFactor(self, term, denom_homo_coeff, param_list, unknown, remaining_terms, equ_degree):
        '''
        In the case where there are enough terms left to be added for not having to start cancelling out high powers of the unknown,
        generates an unknwon dependent factor and multiply the given term by it:

        term (sympy expression) = the term by which to multiply the generated factor
        denom_homo_coeff (list(token)) = coefficients of the polynomial of the denominator of the equation in its homogeneous form (without the considered term)
        param_list (list(sympy.Symbol)) = list of parameters that may appear in the equation
        unknown (sympy.Symbol) = unknown of the equation
        remaining_terms (int) = number of terms yet to be added to the equation after this
        equ_degree (int) = degree of the equation without the considered term
        returns:
            (sympy expression) if possible to put an additional factor
            -1 if not
        '''

        new_factor = self._generateUnknownFactor(param_list, unknown)

        rand = rd.uniform(0, 1)
        if remaining_terms >= (equ_degree+1) - 1 and rand < self.unknown_division_proba:
            term /= new_factor
        elif self._canAddFactor(term, remaining_terms, unknown):
            term *= new_factor
        else:
            return -1
        
        term = self._preventFractionUnificationOrderOverflow(term, denom_homo_coeff, remaining_terms, unknown)

        return term

    def _canAddFactor(self, term, remaining_terms, unknown):
        '''
        Considered the number of terms yet to be added and the current value of the generated term, determines wether this term can be multiplied
        by another unknown dependent factor, meaning that its polynomial order will be incremented:

        term (sympy expression) = the value of this term
        remaining_terms (int) = number of terms yet to be added to the equation after this
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            bool
        '''

        num_order = self._getPolyOrder(numer(term), unknown)
        if remaining_terms < (num_order+1) - 1:
            return False
        else:
            return True
        
    def _preventFractionUnificationOrderOverflow(self, term, denom_homo_coeff, remaining_terms, unknown):
        '''
        Considered that in the homogeneous form, this term will be multiplied by the denominator of the current equation in its homogeneous form,
        compares the resulting order of this term to the maximum acceptable order and divides the term by a certain number of the denominator's
        factors, thus decreasing the resulting order of this term by the same number:

        term (sympy expression) = the value of this term
        denom_homo_coeff (list(token)) = coefficients of the polynomial of the denominator of the equation in its homogeneous form (without the considered term)
        remaining_terms (int) = number of terms yet to be added to the equation after this
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            sympy expression
        '''

        missing_homo_denom_factors = self._getFactorsToCommonDenom(term, denom_homo_coeff, unknown)

        unified_num_order = self._getPolyOrder(numer(term), unknown) + len(missing_homo_denom_factors)
        max_allowed_order = remaining_terms + 1
        order_to_decrease = unified_num_order - max_allowed_order

        if order_to_decrease > 0:
            term = self._decreaseUnifiedNumerOrder(order_to_decrease, term, missing_homo_denom_factors)

        return term

    def _getFactorsToCommonDenom(self, term, denom_homo_coeff, unknown):
        '''
        Considered that in the homogeneous form, this term will be multiplied by the denominator of the current equation in its homogeneous form,
        computes the unknown dependent factors by which the current term will have to be multiplied to be added to the rest of
        the homogeneous equation:

        term (sympy expression) = the value of this term
        denom_homo_coeff (list(token)) = coefficients of the polynomial of the denominator of the equation in its homogeneous form (without the considered term)
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            list(sympy expression)
        '''

        factorized_homo_denom = self._getFactorizedPoly(denom_homo_coeff, unknown)
        expr = denom(term) * factorized_homo_denom
        expr = simplify(expr)

        missing_homo_denom_factors = self._getUnknownDependentFactors(numer(expr), unknown)

        return missing_homo_denom_factors

    def _getFactorizedPoly(self, polynomial_coeffs, unknown):
        '''
        Factorizes the given polynomial:

        polynomial_coeffs (list(token)) = list of the coefficients of the polynomial to be factorized
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            sympy expression
        '''

        polynomial = Poly(polynomial_coeffs, unknown)
        expr = polynomial.as_expr()
        factorized_expr = factor(expr)
        
        return factorized_expr
    
    def _getPolyOrder(self, polynomial_expr, unknown):
        '''
        Computes the order of the given polynomial:

        polynomial_expr (sympy expression) =  polynomial
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            int
        '''

        polynomial = Poly(polynomial_expr, unknown)
        coeffs = polynomial.all_coeffs()

        return len(coeffs) - 1
    
    def _getUnknownDependentFactors(self, polynomial, unknown):
        '''
        Given a polynomial, factorizes it and selects the resulting factors that are unknown dependent:

        polynomial (sympy expression) =  polynomial
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            list(sympy expression)
        '''

        factorized_poly = factor(polynomial)
        unknown_dependent_factors = []

        if factorized_poly.func == Mul:
            factors = factorized_poly.args

            for fact in factors:
                factor_order = self._getPolyOrder(fact, unknown)
                if factor_order > 0:
                    unknown_dependent_factors.append(fact)

        else:
            factor_order = self._getPolyOrder(factorized_poly, unknown)
            if factor_order > 0:
                unknown_dependent_factors.append(factorized_poly)

        return unknown_dependent_factors

    def _decreaseUnifiedNumerOrder(self, order_to_decrease, term, missing_homo_denom_factors):
        '''
        Considered that in the homogeneous form, this term will be multiplied by the denominator of the current equation in its homogeneous form,
        divides the term by factors of the denominator in order to decrease its order in the homogeneous form:

        order_to_decrease (int) =  number of factors of the homogeneous denominator by which the term will be devided
        term (sympy expression) = the value of this term
        missing_homo_denom_factors (list(sympy expression)) = list of the factors by which the term is yet to be divided to be in the homogeneous form
        returns:
            sympy expression
        '''

        for i in range(order_to_decrease):
            index = rd.randint(0, len(missing_homo_denom_factors)-1)
            factor = missing_homo_denom_factors[index]
            term /= factor
            missing_homo_denom_factors.pop(index)

        return term
    
    def _numerAddFactors(self, term, additional_factors_nb, param_list, unknown):
        '''
        Multiplies the term by a certain number of unknown dependent factors:

        term (sympy expression) = the value of this term
        additional_factors_nb (int) = number of factors by which to multiply the term
        param_list (list(sympy.Symbol)) = list of parameters that may appear in the equation
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            sympy expression
        '''

        for i in range(additional_factors_nb):
            new_factor = self._generateUnknownFactor(param_list, unknown)
            term *= new_factor

        return term

    def _complexifyTerm(self, term, denom_homo_coeff, param_list, unknown):
        '''
        Multiply the term by some additional factors if possible according to the generator's parameters,
        and compensate at the denominator with factors of the homogeneous denominator of the equation to keep the order of
        this term in its homogeneous form:

        term (sympy expression) = the value of this term
        denom_homo_coeff (list(token)) = coefficients of the polynomial of the denominator of the equation in its homogeneous form (without the considered term)
        param_list (list(sympy.Symbol)) = list of parameters that may appear in the equation
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            sympy expression
        '''

        numer_factors_nb = len(self._getUnknownDependentFactors(numer(term), unknown))
        denom_factors_nb = len(self._getUnknownDependentFactors(denom(term), unknown))
        tot_factors_nb = numer_factors_nb + denom_factors_nb
        nb_factors = rd.randint(self.add_unknown_factors_range[0], self.add_unknown_factors_range[1])

        additional_factors_nb = nb_factors - tot_factors_nb
        missing_homo_denom_factors = self._getFactorsToCommonDenom(term, denom_homo_coeff, unknown)
        if len(missing_homo_denom_factors) < additional_factors_nb:
            additional_factors_nb = len(missing_homo_denom_factors)

        term = self._numerAddFactors(term, additional_factors_nb, param_list, unknown)
        term = self._decreaseUnifiedNumerOrder(additional_factors_nb, term, missing_homo_denom_factors)

        return term
        
    def _normalizeToCancelGreatestOrder(self, order_coeff, term, denom_homo_coeff, unknown):
        '''
        Computes the number by which to multiply the term for its coefficient in the highest order to be equal to the one given:

        order_coeff (token) = value of the coefficient of the highest order to be reached
        term (sympy expression) = the value of this term
        denom_homo_coeff (list(token)) = coefficients of the polynomial of the denominator of the equation in its homogeneous form (without the considered term)
        unknown (sympy.Symbol) = unknown of the equation
        returns:
            sympy expression
        '''

        unified_term = term * Poly(denom_homo_coeff, unknown).as_expr()
        unified_term = simplify(unified_term)
        coeffs = Poly(numer(unified_term)).all_coeffs()

        normalization_factor = Rational(order_coeff, coeffs[0])
        term *= normalization_factor

        return term"""
