from sympy import *

class Node():
    def __init__(self, is_root=True, is_leaf=False, depth=0, is_correct=True,
                 associated_step=None, solutions=[], parent=None, ex_is_equation=False, ex_expression=[sympify('0')]):
        self.is_root = is_root
        self.is_leaf = is_leaf
        self.depth = depth
        self.is_correct = is_correct
        self.parents_associated_steps = {}
        if associated_step is not None:
            self.parents_associated_steps[parent] = associated_step
        self.solutions = solutions
        self.parents = []
        if parent is not None:
            self.parents.append(parent)
        self.ex_is_equation = ex_is_equation
        self.ex_expression = ex_expression
        self.str_expression = str(Add(*(self.ex_expression), evaluate=False))
        self.children = []

    def equals(self, node):
        """
        Checks whether self is equal to node based on both nodes' string expression:

        node (Node) = Node to compare to self
        returns:
            boolean
        """
        return self.str_expression == node.str_expression
    
    def linkParentToChild(parent, child, associated_step):
        if child not in parent.children:
            parent.children.append(child)
        if parent not in child.parents:
            child.parents.append(parent)
        child.parents_associated_steps[parent] = associated_step

    def isAncestor(self, node, visited_nodes=[]):
        """
        Checks whether self is an ancestor of node:

        node (Node) = Node to compare to self
        visited_nodes (list(Node)) = List of nodes that already have been visited during the algorithm's execution
        returns:
            boolean
        """
        for visited_node in visited_nodes:
            if node.equals(visited_node):
                return False
            
        visited_nodes.append(node)
        for parent in node.parents:
            if parent.equals(self) or self.isAncestor(parent, visited_nodes=visited_nodes):
                return True
        
        return False
    

class ReasoningTree():
    def __init__(self, root=Node(), max_depth=5):
        self.root = root
        self.nodes = [root]
        self.max_depth = max_depth
        self.leaves = []
        self.sol_to_leaves = {}
        if root.is_leaf:
            self.leaves.append(root)
            self.sol_to_leaves[str(root.solutions)] = [root]
        self.last_error_nodes = []

    def find(self, node):
        """
        Searches for an already existing node that is associated to the same expression as node.
        Returns None if none found:

        node (Node) = Node to search
        returns:
            Node
        """
        for tree_node in self.nodes:
            if node.equals(tree_node):
                return tree_node
            
        return None
        
    def appendNode(self, node):
        self.nodes.append(node)
        if node.is_leaf:
            self.leaves.append(node)
            if str(node.solutions) in self.sol_to_leaves:
                self.sol_to_leaves[str(node.solutions)].append(node)
            else:
                self.sol_to_leaves[str(node.solutions)] = [node]

    def findShortestPathFromRoot(self, to_node):
        """
        Computes the shortest path from the root to to_node:

        to_node (Node) = Node to reach
        returns:
            list(dict<"node":Node, "next step":Step, "next node":Node>)
        """
        visited_nodes = {}
        unvisited_neighbors = {0: [{"node": to_node,"child": None,"step": None}]}

        #compute reverse paths from to-node to root for efficiency reasons
        while self.root.str_expression not in visited_nodes:
            next_shortest_path_length = min(unvisited_neighbors.keys())
            visited_node = unvisited_neighbors[next_shortest_path_length].pop(0)
            if len(unvisited_neighbors[next_shortest_path_length]) == 0:
                unvisited_neighbors.pop(next_shortest_path_length)

            visited_node["path length"] = next_shortest_path_length
            key = visited_node["node"].str_expression
            visited_nodes[key] = visited_node

            for neighbor in visited_node["node"].parents:
                if neighbor.str_expression not in visited_nodes:
                    step = visited_node["node"].parents_associated_steps[neighbor]
                    path_length = next_shortest_path_length + step.weight
                    if path_length not in unvisited_neighbors:
                        unvisited_neighbors[path_length] = []
                    unvisited_neighbors[path_length].append({"node": neighbor,"child": visited_node["node"],"step": step})

        next_node = visited_nodes[self.root.str_expression]
        shortest_path = []
        #retrieve shortest path 
        while not next_node["node"].is_leaf:
            shortest_path.append({"node": next_node["node"], "next step": next_node["step"], "next node": next_node["child"]})
            next_node_key = next_node["child"].str_expression
            next_node = visited_nodes[next_node_key]

        return shortest_path