import sys
import numpy as np
import matplotlib.pyplot as plt
import json
import time
from sympy import *

from corrector import *
from generator import *

def getGenerator(ex_type, config, seed=None):
    config_file = None
    if ex_type == "fraction_simplification":
        corrector = FracSimpCorrector()
        generator = FracSimpGenerator(config_file, corrector, config=config, seed=seed)
    elif ex_type == "first_degree_equation":
        corrector = FirstDegEquCorrector()
        generator = FirstDegEquGenerator(config_file, corrector, config=config, seed=seed)

    return generator

def generateData(ex_type, config):
    """
    Computes timings, mean time and error of exercise generation:

    ex_type (string) = Type of exercise to generate  (fraction_simplification or first_degree_equation)
    config (dict) = Configuration for the exercise generator
    returns:
        dict("times":list(float), "mean":float, "error":float)
    """
    times = timeGeneration(ex_type, config)
    batch_means = []
    for batch in times:
        batch_means.append(np.mean(batch))

    error = np.std(batch_means, ddof=1) / np.sqrt(len(batch_means))
    
    data = {"times": times, "mean": np.mean(batch_means), "error": error}
    return data

def timeGeneration(ex_type, config):
    """
    Computes timings of exercise generation:

    ex_type (string) = Type of exercise to generate  (fraction_simplification or first_degree_equation)
    config (dict) = Configuration for the exercise generator
    returns:
        list(float)
    """
    times = []
    for batch in range(nb_batch):
        batch_times = []
        generator = getGenerator(ex_type, config, seed=batch + (test*nb_batch))
        for j in range(nb_in_batch):
            start_time = time.time()
            exercise = generator.generateEx()
            end_time = time.time()
            batch_times += [end_time - start_time]
            print(j, end_time - start_time)
        
        times.append(batch_times)
        print("Generated " + str(batch+1) + " out of " + str(nb_batch) + " batches")
        print("Mean time per exercise: " + str(round(np.mean(batch_times))) + " second")
    
    return times
    

if len(sys.argv) != 1 and (len(sys.argv) != 2 or sys.argv[1] != "-graphOnly"):
    sys.stdout.write("./testGenerators.py [-graphOnly]\n")
    sys.stdout.write("     where -graphOnly if you want only to rebuild the graph based on already computed data.\n")
    exit("Bad arguments\n")

default_config = {
    "nb_param_range": [0, 2],
    "possible_params": ["a", "b", "c", "d", "x", "y", "z", "u", "v", "w"],

    "max_int_value": 5,

    "frac_simplification": {

        "answer_num_complexity_range": [0, 3],
        "answer_denom_complexity_range": [0, 3],

        "nb_factor_range": [1, 3],
        "factor_complexity_range": [0, 0]
    },

    "first_degree_equation": {
        
        "possible_unknowns": ["x"],
        "fraction_token_proba": 0.25,
        "complex_factor_proba": 0.5,
        "add_unknown_terms_range": [0, 2],
        "add_unknown_factors_range": [0, 0],
        "unknown_division_proba": 0
    }
}

graph_only = False
if len(sys.argv) == 2:
    graph_only = True

nb_batch = 5
nb_in_batch = 20
test = 0
data_file = open('testCorrectorsData.json')
whole_data = json.load(data_file)
data_file.close()
                    
#Generate fraction_simplification data
if not graph_only:  #generate new data
    if "fraction_simplification" not in whole_data:
        whole_data["fraction_simplification"] = {}
    ex_type = "fraction_simplification"
    print("Generate " + ex_type + " exercises")

    #Generate simple_data
    config = default_config.copy()
    config["nb_param_range"] = [0, 0]
    config["frac_simplification"]["answer_num_complexity_range"] = [0, 0]
    config["frac_simplification"]["answer_denom_complexity_range"] = [0, 0]
    config["frac_simplification"]["nb_factor_range"] = [2, 2]
    print("Generate simple_data")
    simple_data = generateData(ex_type, config)
    whole_data["fraction_simplification"]["simple_data"] = simple_data
    with open('testCorrectorsData.json', 'w') as fp:
        json.dump(whole_data, fp)
    print()

    #Generate mid_data
    test += 1
    config["nb_param_range"] = [1, 1]
    config["frac_simplification"]["answer_num_complexity_range"] = [0, 1]
    config["frac_simplification"]["answer_denom_complexity_range"] = [0, 1]
    config["frac_simplification"]["factor_complexity_range"] = [1, 1]
    print("Generate mid_data")
    mid_data = generateData(ex_type, config)
    whole_data["fraction_simplification"]["mid_data"] = mid_data
    with open('testCorrectorsData.json', 'w') as fp:
        json.dump(whole_data, fp)
    print()

    #Generate complex_data
    test += 1
    config["nb_param_range"] = [2, 2]
    config["frac_simplification"]["answer_num_complexity_range"] = [1, 2]
    config["frac_simplification"]["answer_denom_complexity_range"] = [1, 2]
    config["frac_simplification"]["factor_complexity_range"] = [2, 2]
    print("Generate complex_data")
    complex_data = generateData(ex_type, config)
    whole_data["fraction_simplification"]["complex_data"] = complex_data
    with open('testCorrectorsData.json', 'w') as fp:
        json.dump(whole_data, fp)
    print()
    print()

simple_data = whole_data["fraction_simplification"]["simple_data"]
mid_data = whole_data["fraction_simplification"]["mid_data"]
complex_data = whole_data["fraction_simplification"]["complex_data"]
names = ["simple", "mid-complex", "complex"]
values = [simple_data["mean"], mid_data["mean"], complex_data["mean"]]
errors = [simple_data["error"], mid_data["error"], complex_data["error"]]

#Generate graphs off mean times for simple, mid-complex and complex exercises
plt.xlabel('Exercise complexity')
plt.ylabel('Average generation time (sec)')
plt.bar(names, values, color='b')
for i in range(len(names)):
        plt.text(names[i], values[i]+1, round(values[i]), ha = 'left')
plt.errorbar(names, values, yerr=errors, fmt="None", ecolor='r')
plt.title("Average time for generating exercises")
plt.savefig('graphs/Corrector/FS_mean_generation_time.png', bbox_inches='tight')
#plt.show()
plt.clf()

values = []
categories = [simple_data, mid_data, complex_data]
for elem in categories:
    data = []
    for batch in elem["times"]:
        data += batch
    values.append(data)

#Generate boxplots for simple, mid-complex and complex exercises
for i in range(len(names)):
    plt.ylabel('Generation time (sec)')
    bplot = plt.boxplot(values[i], showmeans=True, meanline=True)
    plt.xticks([])
    plt.legend([bplot['medians'][0], bplot['means'][0]], ['median', 'mean'])
    plt.title("Boxplot of generation time for " + names[i] + " exercises")
    plt.savefig('graphs/Corrector/FS_generation_time_' + names[i] + '.png', bbox_inches='tight')
    plt.clf()


#Generate first_degree_equation data
if not graph_only:  #generate new data
    if "first_degree_equation" not in whole_data:
        whole_data["first_degree_equation"] = {}
    ex_type = "first_degree_equation"
    print("Generate " + ex_type + " exercises")

    #Generate simple_data
    test += 1
    config = default_config.copy()
    config["nb_param_range"] = [0, 0]
    config["first_degree_equation"]["add_unknown_terms_range"] = [0, 0]
    print("Generate simple_data")
    simple_data = generateData(ex_type, config)
    whole_data["first_degree_equation"]["simple_data"] = simple_data
    with open('testCorrectorsData.json', 'w') as fp:
        json.dump(whole_data, fp)
    print()

    #Generate mid_data
    test += 1
    config["first_degree_equation"]["fraction_token_proba"] = 0.75
    config["first_degree_equation"]["complex_factor_proba"] = 0.75
    config["first_degree_equation"]["add_unknown_terms_range"] = [1, 1]
    print("Generate mid_data")
    mid_data = generateData(ex_type, config)
    whole_data["first_degree_equation"]["mid_data"] = mid_data
    with open('testCorrectorsData.json', 'w') as fp:
        json.dump(whole_data, fp)
    print()

    #Generate complex_data
    test += 1
    config["first_degree_equation"]["add_unknown_terms_range"] = [1, 2]
    config["first_degree_equation"]["add_unknown_factors_range"] = [1, 1]
    print("Generate complex_data")
    complex_data = generateData(ex_type, config)
    whole_data["first_degree_equation"]["complex_data"] = complex_data
    with open('testCorrectorsData.json', 'w') as fp:
        json.dump(whole_data, fp)
    print()
    print()

simple_data = whole_data["first_degree_equation"]["simple_data"]
mid_data = whole_data["first_degree_equation"]["mid_data"]
complex_data = whole_data["first_degree_equation"]["complex_data"]
names = ["simple", "mid-complex", "complex"]
values = [simple_data["mean"], mid_data["mean"], complex_data["mean"]]
errors = [simple_data["error"], mid_data["error"], complex_data["error"]]

#Generate graphs off mean times for simple, mid-complex and complex exercises
plt.xlabel('Exercise complexity')
plt.ylabel('Average generation time (sec)')
plt.bar(names, values, color='b')
for i in range(len(names)):
        plt.text(names[i], values[i]+1, round(values[i]), ha = 'left')
plt.errorbar(names, values, yerr=errors, fmt="None", ecolor='r')
plt.title("Average time for generating exercises")
plt.savefig('graphs/Corrector/FDE_mean_generation_time.png', bbox_inches='tight')
#plt.show()
plt.clf()

values = []
categories = [simple_data, mid_data, complex_data]
for elem in categories:
    data = []
    for batch in elem["times"]:
        data += batch
    values.append(data)

#Generate boxplots for simple, mid-complex and complex exercises
for i in range(len(names)):
    plt.ylabel('Generation time (sec)')
    bplot = plt.boxplot(values[i], showmeans=True, meanline=True)
    plt.xticks([])
    plt.legend([bplot['medians'][0], bplot['means'][0]], ['median', 'mean'])
    plt.title("Boxplot of generation time for " + names[i] + " exercises")
    plt.savefig('graphs/Corrector/FDE_generation_time_' + names[i] + '.png', bbox_inches='tight')
    plt.clf()