import sys
import numpy as np
import matplotlib.pyplot as plt
from sympy import *
from enum import Enum

from corrector import *
from generator import *

def generateExercises():
    exercises = []
    for batch in range(nb_batch):
        batch_exercises = []
        generator = getGenerator(seed=batch + (test*nb_batch))
        for j in range(nb_in_batch):
            exercise = generator.generateEx()
            batch_exercises += [exercise]
        
        exercises.append(batch_exercises)
    
    return exercises

def getGenerator(seed=None):
    config_file = None
    if mode == Mode.FRACTION_SIMPLIFICATION:
        corrector = FracSimpCorrector()
        generator = FracSimpGenerator(config_file, corrector, config=config, seed=seed)
    elif mode == Mode.FIRST_DEGREE_EQUATION:
        corrector = FirstDegEquCorrector()
        generator = FirstDegEquGenerator(config_file, corrector, config=config, seed=seed)

    return generator

def FSTest(config, nb_param_range_values):
    """
    Performs tests on nb_param_range, answer_num_complexity_range, answer_denom_complexity_range
    and factor_complexity_range for fracrion simplification exercises:

    config (dict) = Configuration for the exercise generator
    nb_param_range_values (list(list(int, int))) = Values for both bounds in nb_param_range
    returns:
        answer_diversities (list(dict<"values":list(list(int, int)), "means":list(float), "errors":list(float), "diversities":list(float), "diversities_pmfs":list(numpy.array(float))>))
        factor_diversities (list(dict<"values":list(list(int, int)), "means":list(float), "errors":list(float), "diversities":list(float), "diversities_pmfs":list(numpy.array(float))>))
    """
    config["frac_simplification"]["nb_factor_range"] = [1, 1]

    answer_diversities = nb_param_range_values.copy()
    factor_diversities = nb_param_range_values.copy()

    for nb_param in nb_param_range_values:
        config["nb_param_range"] = [nb_param, nb_param]

        answer_diversities[nb_param] = {}
        answer_diversities[nb_param]["values"] = []
        answer_diversities[nb_param]["means"] = []
        answer_diversities[nb_param]["errors"] = []
        answer_diversities[nb_param]["diversities"] = []
        answer_diversities[nb_param]["diversities_pmfs"] = []
        factor_diversities[nb_param] = {}
        factor_diversities[nb_param]["values"] = []
        factor_diversities[nb_param]["means"] = []
        factor_diversities[nb_param]["errors"] = []
        factor_diversities[nb_param]["diversities"] = []
        factor_diversities[nb_param]["diversities_pmfs"] = []

        for complexity in range(4):
            config["frac_simplification"]["answer_num_complexity_range"] = [complexity, complexity]
            config["frac_simplification"]["answer_denom_complexity_range"] = [complexity, complexity]
            config["frac_simplification"]["factor_complexity_range"] = [complexity, complexity]
            exercises = generateExercises()

            answer_mean, answer_error = FSGetAnswerStats(exercises)
            factor_mean, factor_error = FSGetGFactorStats(exercises)
            ex_list = []
            for batch in exercises:
                ex_list += batch
            answer_diversity, answer_pmf = getExDiversity(ex_list, "answer_num/denom_complexity_range = " + str([complexity, complexity]) + ", nb_param_range = " + str([nb_param, nb_param]))
            factor_diversity, factor_pmf = getFactorDiversity(ex_list, "factor_complexity_range = " + str([complexity, complexity]) + ", nb_param_range = " + str([nb_param, nb_param]))

            answer_diversities[nb_param]["values"].append([complexity, complexity])
            answer_diversities[nb_param]["means"].append(answer_mean)
            answer_diversities[nb_param]["errors"].append(answer_error)
            answer_diversities[nb_param]["diversities"].append(answer_diversity)
            answer_diversities[nb_param]["diversities_pmfs"].append(answer_pmf)

            factor_diversities[nb_param]["values"].append([complexity, complexity])
            factor_diversities[nb_param]["means"].append(factor_mean)
            factor_diversities[nb_param]["errors"].append(factor_error)
            factor_diversities[nb_param]["diversities"].append(factor_diversity)
            factor_diversities[nb_param]["diversities_pmfs"].append(factor_pmf)

    return answer_diversities, factor_diversities

def FSGetAnswerStats(exercises):
    """
    Computes mean value and error of answer complexities for fraction simplification exercises:

    exercises (list(list(Exercise))) = list of batches of exercise
    returns:
        mean (float)
        error (float)
    """
    batch_means = []
    for batch in exercises:
        measured_complexities = []
        for ex in batch:
            measured_complexity = FSGetAnswerComplexity(ex.answers[0])
            measured_complexities.append(measured_complexity)
        
        batch_means.append(np.mean(measured_complexities))

    error = np.std(batch_means, ddof=1) / np.sqrt(len(batch_means))
    
    return np.mean(batch_means), error

def FSGetAnswerComplexity(answer):
    num_complexity = getExpressionComplexity(numer(answer))
    denom_complexity = getExpressionComplexity(denom(answer))

    return num_complexity + denom_complexity

def getExpressionComplexity(expression):
    if len(expression.args) == 0:
        return 0
    elif len(expression.args)==2 and expression.func==Mul and -1 in expression.args:
        return 0
    
    complexity = len(expression.args) - 1
    for arg in expression.args:
        complexity += getExpressionComplexity(arg)

    return complexity

def FSGetGFactorStats(exercises):
    """
    Computes mean value and error of factors complexities for fraction simplification exercises:

    exercises (list(list(Exercise))) = list of batches of exercise
    returns:
        mean (float)
        error (float)
    """
    batch_means = []
    for batch in exercises:
        measured_complexities = []
        for ex in batch:
            measured_complexity = getExpressionComplexity(ex.factors[0])
            measured_complexities.append(measured_complexity)
        
        batch_means.append(np.mean(measured_complexities))

    error = np.std(batch_means, ddof=1) / np.sqrt(len(batch_means))
    
    return np.mean(batch_means), error

def FDETestProba(param, checkFunction, config):
    """
    Performs test on probability parameter param for first degree equation exercises:

    param (string) = Name of the parameter to test
    checkFunction (function) = Function that should be used to count the number of occurence of elements of interest
    returns:
        values (list(float))
        means (list(float))
        errors (list(float))
        diversities (list(float))
        diversities_pmfs (list(numpy.array(float)))
    """
    config["nb_param_range"] = [0, 0]           #Introducing parameters makes the exercises harder to check. Plus, in most cases, paramters won't be used in this kind of exercises.
    values = []
    means = []
    errors = []
    diversities = []
    diversities_pmfs = []

    for i in range(nb_param_values+1):
        value = round(1./nb_param_values * i, 3)
        config["first_degree_equation"][param] = value
        exercises = generateExercises()

        mean, error = FDEGetProbaStats(exercises, checkFunction)
        ex_list = []
        for batch in exercises:
            ex_list += batch
        diversity, pmf = getExDiversity(ex_list, param + " = " + str(value))

        values.append(value)
        means.append(mean)
        errors.append(error)
        diversities.append(diversity)
        diversities_pmfs.append(pmf)

    return values, means, errors, diversities, diversities_pmfs

def FDEGetProbaStats(exercises, checkFunction):
    """
    Computes mean value and error number of occurence of elements of interest for first degree equation exercises:

    exercises (list(list(Exercise))) = list of batches of exercise
    checkFunction (function) = Function that should be used to count the number of occurence of elements of interest
    returns:
        mean (float)
        error (float)
    """
    batch_means = []
    for batch in exercises:
        ratios = []
        for ex in batch:
            nb_element, nb_specific_element = checkFunction(Add(ex.ex_expr[0], ex.ex_expr[1], evaluate=False), ex.unknowns[0])
            ratios.append(nb_specific_element/nb_element)
        
        batch_means.append(np.mean(ratios))

    error = np.std(batch_means, ddof=1) / np.sqrt(len(batch_means))
    
    return np.mean(batch_means), error

def getTokenAndFractionTokenNb(expr, unknown):
    """
    Counts tokens and fraction tokens in expr:

    expr (sympy expression) = Expression to analyze
    unknown (sympy.Symbol) = Unknown of the equation
    returns:
        nb_token (int)
        nb_token_fraction (int)
    """
    nb_token = 0
    nb_token_fraction = 0
    if len(expr.args) == 0:
        if expr == unknown:
            return nb_token, nb_token_fraction
        elif expr.func is Rational or expr is S.Half:
            return nb_token+1, nb_token_fraction+1
        else:
            return nb_token+1, nb_token_fraction
        
    else:
        for arg in expr.args:
            t, tf = getTokenAndFractionTokenNb(arg, unknown)
            nb_token += t
            nb_token_fraction += tf
        
        return nb_token, nb_token_fraction
    
def getFactorAndComplexFactorNb(expr, unknown):
    """
    Counts factors and complex factors in expr:

    expr (sympy expression) = Expression to analyze
    unknown (sympy.Symbol) = Unknown of the equation
    returns:
        nb_factors (int)
        nb_complex_factors (int)
    """
    nb_factors = 0
    nb_complex_factors = 0
    str_unknown = str(unknown)
    
    for side in expr.args: #expr is sum of left and right hand sides (thus expr.args are both sides)
        str_side = str(side)
        index = str_side.find(str_unknown)
        if index != -1: #Side of equation is not just a constant which would not include any unknown dependent factor
            index = str_side.find(str_unknown, index+1)
            if index == -1: #Side is just composed of one term composed of one unknown dependent factor
                nb_factors += 1
                nb_complex_factors += isComplexFactor(side, unknown, False)
            else:   #Side is composed of several terms composed of at most one unknown dependent factor (could be a constant)
                for term in side.args:
                    nb_factors += (str(term).find(str_unknown) != -1)
                    nb_complex_factors += isComplexFactor(term, unknown, False)
    
    return nb_factors, nb_complex_factors

def isComplexFactor(factor, unknown, includes_mul):
    """
    The input factor should be unknown dependent, of the form either a*x+b (a could be one and b 0) of a*(b*x+c)+d (b could be one and c and d 0)
    thus if the factor were a complex factor, it would contain a product (a * (b*x+c)) (includes_mul = True) whose one arg contains a sum
    (b*x + c) whose one arg is term dependent (b*x). 
    """
    if factor == unknown:
        return False

    elif factor.func is Add:
        for arg in factor.args:
            if str(arg).find(str(unknown)) != -1:
                if includes_mul:
                    return True
                else:
                    return isComplexFactor(arg, unknown, includes_mul)
                
    elif factor.func is Mul:
        for arg in factor.args:
            if str(arg).find(str(unknown)) != -1:
                return isComplexFactor(arg, unknown, True)

    return False

def getExDiversity(exercises, param_value_str):
    """
    Compute the exercises' diversity and its pmf:

    exercises (list(Exercise)) = Exercises to analyze
    param_value_str (string) = Name of the parameter under analysis (for print)
    returns:
        diversity (float)
        pmf (numpy.array(float))
    """
    answers = {}
    for ex in exercises:
        str_ex = str(Add(*ex.ex_expr, evaluate=False))
        str_ans = str(ex.answers)
        if str_ans in answers:
            answers[str_ans][0] += 1
            if str_ex in answers[str_ans][1]:
                answers[str_ans][1][str_ex] += 1
            else:
                answers[str_ans][1][str_ex] = 1
        else:
            answers[str_ans] = [1, {str_ex: 1}]

    sys.stdout.write("With the parameter " + param_value_str + ", " + str(len(answers)) + "/" + str(len(exercises)) + " generated exercises with different answers\n")
    diversity = len(answers)/len(exercises)

    values = list(answers.values())
    max_elem = max(zip([elem[0] for elem in answers.values()], answers.keys()))
    max_value = max_elem[0]
    max_answer = max_elem[1]
    pmf = np.zeros(max_value)
    for value in values:
        pmf[value[0]-1] += 1
    pmf /= len(values)

    #print("Most common answer:", max_answer)
    #print("Associated exercises:", answers[max_answer][1])

    return diversity, pmf

def getFactorDiversity(exercises, param_value_str):
    """
    Compute the exercises' factors diversity and its pmf:

    exercises (list(Exercise)) = Exercises to analyze
    param_value_str (string) = Name of the parameter under analysis (for print)
    returns:
        diversity (float)
        pmf (numpy.array(float))
    """
    factors = {}
    for ex in exercises:
        str_ex = str(Add(*ex.ex_expr, evaluate=False))
        str_factor = str(ex.factors[0])
        if str_factor in factors:
            factors[str_factor][0] += 1
            if str_ex in factors[str_factor][1]:
                factors[str_factor][1][str_ex] += 1
            else:
                factors[str_factor][1][str_ex] = 1
        else:
            factors[str_factor] = [1, {str_ex: 1}]

    sys.stdout.write("With the parameter " + param_value_str + ", " + str(len(factors)) + "/" + str(len(exercises)) + " generated factors with different expressions\n")
    diversity = len(factors)/len(exercises)

    values = list(factors.values())
    max_elem = max(zip([elem[0] for elem in factors.values()], factors.keys()))
    max_value = max_elem[0]
    max_factor = max_elem[1]
    pmf = np.zeros(max_value)
    for value in values:
        pmf[value[0]-1] += 1
    pmf /= len(values)

    #print(factors.keys())
    #print("Most common answer:", max_factor)
    #print("Associated exercises:", factors[max_factor][1])

    return diversity, pmf

def roundToFirstSignificantDigit(nb, min_digits):
    if nb == 0.:
        return int(0)
    
    i = min_digits
    while round(nb, i) == 0:
        i += 1
    
    return round(nb, i)

class Mode(Enum):
    FRACTION_SIMPLIFICATION = 1
    FIRST_DEGREE_EQUATION = 2

OPTIONS = {}
OPTIONS["fraction_simplification"] = Mode.FRACTION_SIMPLIFICATION
OPTIONS["first_degree_equation"] = Mode.FIRST_DEGREE_EQUATION

if len(sys.argv) != 2 or not sys.argv[1] in OPTIONS :
    sys.stdout.write("./testGenerators.py <ex_type>\n")
    sys.stdout.write("     where <ex_type> can be fraction_simplification, first_degree_equation\n")
    exit("Bad arguments\n")

mode = OPTIONS[sys.argv[1]]

config = {
    "nb_param_range": [0, 2],
    "possible_params": ["a", "b", "c", "d", "x", "y", "z", "u", "v", "w"],

    "max_int_value": 5,

    "frac_simplification": {

        "answer_num_complexity_range": [0, 3],
        "answer_denom_complexity_range": [0, 3],

        "nb_factor_range": [1, 3],
        "factor_complexity_range": [0, 0]
    },

    "first_degree_equation": {
        
        "possible_unknowns": ["x"],
        "fraction_token_proba": 0.25,
        "complex_factor_proba": 0.5,
        "add_unknown_terms_range": [0, 2],
        "add_unknown_factors_range": [0, 0],
        "unknown_division_proba": 0
    }
}

nb_batch = 20
nb_in_batch = 50
nb_param_values = 10

if mode == Mode.FRACTION_SIMPLIFICATION:
    default_config = config

    color = 'b'
    ecolor = 'r'

    means_sym = 'o'
    diversities_sym = '^'

    #test fraction_token_proba
    test = 0
    nb_param_range_values = list(range(3))
    multi_plot_colors = ['c', 'm', 'y']
    answer_diversity, factor_diversity = FSTest(config, nb_param_range_values)

    values = []
    for elem in answer_diversity[0]["values"]:
        values.append(elem[0])

    #Generate graph of average measured answer complexity
    plt.xlabel('Answer numerator/denominator complexity')
    plt.ylabel('Average measured answer complexity')
    legend_messages = []
    for i in nb_param_range_values:
        plt.plot(values, answer_diversity[i]["means"], c=multi_plot_colors[i], marker=means_sym, label='answer_complexity mean')
        plt.errorbar(values, answer_diversity[i]["means"], yerr=answer_diversity[i]["errors"], ecolor=ecolor)
        legend_messages.append("nb_param_range=" + str([i, i]))
    plt.legend(legend_messages, loc="upper left")
    plt.title("Measured answer complexity as a function of\nanswer_num_complexity_range and answer_denom_complexity_range values")
    plt.savefig('graphs/Fraction_simplification/Answer_complexity/FS_answer_complexity_mean.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graph of exercise diversity
    plt.ylim(-0.1, 1.1)
    plt.xlabel('Answer numerator/denominator complexity')
    plt.ylabel('Exercise diversity')
    legend_messages = []
    for i in nb_param_range_values:
        plt.plot(values, answer_diversity[i]["diversities"], c=multi_plot_colors[i], marker=diversities_sym, label='answer_complexity diversity')
        legend_messages.append("nb_param_range=" + str([i, i]))
    plt.legend(legend_messages, loc="lower right")
    plt.title("Exercise diversity as a function of\nanswer_num_complexity_range and answer_denom_complexity_range values")
    plt.savefig('graphs/Fraction_simplification/Answer_complexity/FS_answer_complexity_diversity.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graphs of exercise diversities pmfs
    for nb_param in nb_param_range_values:
        for i in range(len(answer_diversity[nb_param]["diversities_pmfs"])):
            pmf = answer_diversity[nb_param]["diversities_pmfs"][i]
            x = np.arange(1, len(pmf)+1)
            trimed_pmf = []
            trimed_x = []
            for j in range(len(pmf)):
                if pmf[j] != 0:
                    trimed_pmf.append(pmf[j])
                    trimed_x.append(x[j])
            plt.ylim(0, 1.1)
            plt.xlabel('Nb of exercises sharing the common answer')
            plt.ylabel('Percentage')
            plt.scatter(trimed_x, trimed_pmf, c=color)
            plt.vlines(trimed_x, 0, trimed_pmf, colors=color, linestyles='-', lw=2)
            if trimed_x[-1] < 15:
                step = 1
            elif trimed_x[-1] >= 15 and trimed_x[-1] < 75:
                step = 5
            elif trimed_x[-1] >= 75 and trimed_x[-1] < 150:
                step = 10
            else:
                step = 20
            xi = list(range(0, trimed_x[-1]+1, step))
            plt.xticks(xi, xi)
            plt.title('Percentage of answers that are shared by multiple exercises for\nanswer_num_complexity_range and answer_denom_complexity_range=' + str(answer_diversity[nb_param]["values"][i]) + "\nand nb_param_range=" + str([nb_param, nb_param]))      
            plt.savefig('graphs/Fraction_simplification/Answer_complexity/sharing_ex_proba/'+str(nb_param)+'_param/FS_answer_num-denom_complexity_range' + str(answer_diversity[nb_param]["values"][i]) + 'sharing_ex_proba_with_'+str(nb_param)+'_param.png', bbox_inches='tight')
            plt.clf()

    values = []
    for elem in factor_diversity[0]["values"]:
        values.append(elem[0])

    #Generate graph of average measured factor complexity
    plt.xlabel('Factor complexity')
    plt.ylabel('Average measured factor complexity')
    legend_messages = []
    for i in nb_param_range_values:
        plt.plot(values, factor_diversity[i]["means"], c=multi_plot_colors[i], marker=means_sym, label='factor_complexity mean')
        plt.errorbar(values, factor_diversity[i]["means"], yerr=factor_diversity[i]["errors"], ecolor=ecolor)
        legend_messages.append("nb_param_range=" + str([i, i]))
    plt.legend(legend_messages, loc="upper left")
    plt.title("Measured factor complexity as a function of\nfactor_complexity_range's value")
    plt.savefig('graphs/Fraction_simplification/Factor_complexity/FS_factor_complexity_mean.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graph of average measured factor diversity
    plt.ylim(-0.1, 1.1)
    plt.xlabel('Factor complexity')
    plt.ylabel('Factor diversity')
    legend_messages = []
    for i in nb_param_range_values:
        plt.plot(values, factor_diversity[i]["diversities"], c=multi_plot_colors[i], marker=diversities_sym, label='factor_complexity diversity')
        legend_messages.append("nb_param_range=" + str([i, i]))
    plt.legend(legend_messages, loc="upper left")
    plt.title("Factor diversity as a function of\nfactor_complexity_range's value")
    plt.savefig('graphs/Fraction_simplification/Factor_complexity/FS_factor_complexity_diversity.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graphs of factor diversities pmfs
    for nb_param in nb_param_range_values:
        for i in range(len(factor_diversity[nb_param]["diversities_pmfs"])):
            pmf = factor_diversity[nb_param]["diversities_pmfs"][i]
            x = np.arange(1, len(pmf)+1)
            trimed_pmf = []
            trimed_x = []
            for j in range(len(pmf)):
                if pmf[j] != 0:
                    trimed_pmf.append(pmf[j])
                    trimed_x.append(x[j])
            plt.ylim(0, 1.1)
            plt.xlabel('Nb of factors sharing the common value')
            plt.ylabel('Percentage')
            plt.scatter(trimed_x, trimed_pmf, c=color)
            plt.vlines(trimed_x, 0, trimed_pmf, colors=color, linestyles='-', lw=2)
            if trimed_x[-1] < 15:
                step = 1
            elif trimed_x[-1] >= 15 and trimed_x[-1] < 75:
                step = 5
            elif trimed_x[-1] >= 75 and trimed_x[-1] < 150:
                step = 10
            else:
                step = 20
            xi = list(range(0, trimed_x[-1]+1, step))
            plt.xticks(xi, xi)
            plt.title('Percentage of factor values that are shared by multiple factors for\nfactor_complexity_range=' + str(factor_diversity[nb_param]["values"][i]) + " and nb_param_range=" + str([nb_param, nb_param]))     
            plt.savefig('graphs/Fraction_simplification/Factor_complexity/sharing_factor_proba/'+str(nb_param)+'_param/FS_factor_complexity_range' + str(factor_diversity[nb_param]["values"][i]) + 'sharing_factor_proba_with_'+str(nb_param)+'_param.png', bbox_inches='tight')
            plt.clf()

    config = default_config

elif mode == Mode.FIRST_DEGREE_EQUATION:
    default_config = config

    color = 'b'
    ecolor = 'r'
    means_sym = 'o'
    diversities_sym = '^'

    #test fraction_token_proba
    test = 0
    values, means, errors, diversities, diversities_pmfs = FDETestProba("fraction_token_proba", getTokenAndFractionTokenNb, config)

    #Generate graph of average measured fraction tokens rate
    plt.xlabel('Fraction token probability')
    plt.ylabel('Average fraction tokens rate')
    plt.plot(values, means, c=color, marker=means_sym, label='fraction_token_proba mean')
    plt.errorbar(values, means, yerr=errors, ecolor=ecolor)
    plt.title("Measured fraction token occurence rates as a function of\nfraction_token_proba's value")
    plt.savefig('graphs/First_degree_equation/Fraction_token_probability/FDE_fraction_token_proba_mean.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graph of exercise diversity
    plt.ylim(-0.1, 1.1)
    plt.xlabel('Fraction token probability')
    plt.ylabel('Exercise diversity')
    plt.plot(values, diversities, c=color, marker=diversities_sym, label='fraction_token_proba diversity')
    plt.title("Exercise diversity as a function of\nfraction_token_proba's value")
    plt.savefig('graphs/First_degree_equation/Fraction_token_probability/FDE_fraction_token_proba_diversity.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graphs of exercise diversities pmfs
    for i in range(len(diversities_pmfs)):
        pmf = diversities_pmfs[i]
        x = np.arange(1, len(pmf)+1)
        trimed_pmf = []
        trimed_x = []
        for j in range(len(pmf)):
            if pmf[j] != 0:
                trimed_pmf.append(pmf[j])
                trimed_x.append(x[j])
        plt.ylim(0, 1.1)
        plt.xlabel('Nb of exercises sharing the common answer')
        plt.ylabel('Percentage')
        plt.scatter(trimed_x, trimed_pmf, c=color)
        plt.vlines(trimed_x, 0, trimed_pmf, colors=color, linestyles='-', lw=2)
        if trimed_x[-1] < 15:
            step = 1
        elif trimed_x[-1] >= 15 and trimed_x[-1] < 75:
            step = 5
        elif trimed_x[-1] >= 75 and trimed_x[-1] < 150:
            step = 10
        else:
            step = 20
        xi = list(range(0, trimed_x[-1]+1, step))
        plt.xticks(xi, xi)
        plt.title('Percentage of answers that are shared by multiple\nexercises for fraction_token_proba=' + str(values[i]))
                  
        plt.savefig('graphs/First_degree_equation/Fraction_token_probability/sharing_ex_proba/FDE_fraction_token_proba_' + str(values[i]) + 'sharing_ex_proba.png', bbox_inches='tight')
        plt.clf()

    #test complex_factor_proba
    test += 1
    values, means, errors, diversities, diversities_pmfs = FDETestProba("complex_factor_proba", getFactorAndComplexFactorNb, config)

    #Generate graph of average measured complex factor rate
    plt.xlabel('Complex factor probability')
    plt.ylabel('Average complex factor rate')
    plt.plot(values, means, c=color, marker=means_sym, label='complex_factor_proba mean')
    plt.errorbar(values, means, yerr=errors, ecolor=ecolor)
    plt.title("Measured complex factor occurence rates as a function of\ncomplex_factor_proba's value")
    plt.savefig('graphs/First_degree_equation/Complex_factor_probability/FDE_complex_factor_proba_mean.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graph of exercise diversity
    plt.ylim(-0.1, 1.1)
    plt.xlabel('Complex factor probability')
    plt.ylabel('Exercise diversity')
    plt.plot(values, diversities, c=color, marker=diversities_sym, label='complex_factor_proba diversity')
    plt.title("Exercise diversity as a function of\ncomplex_factor_proba's value")
    plt.savefig('graphs/First_degree_equation/Complex_factor_probability/FDE_complex_factor_proba_diversity.png', bbox_inches='tight')
    #plt.show()
    plt.clf()

    #Generate graphs of exercise diversities pmfs
    for i in range(len(diversities_pmfs)):
        pmf = diversities_pmfs[i]
        x = np.arange(1, len(pmf)+1)
        trimed_pmf = []
        trimed_x = []
        for j in range(len(pmf)):
            if pmf[j] != 0:
                trimed_pmf.append(pmf[j])
                trimed_x.append(x[j])
        plt.ylim(0, 1.1)
        plt.xlabel('Nb of exercises sharing the common answer')
        plt.ylabel('Percentage')
        plt.scatter(trimed_x, trimed_pmf, c=color)
        plt.vlines(trimed_x, 0, trimed_pmf, colors=color, linestyles='-', lw=2)
        if trimed_x[-1] < 15:
            step = 1
        elif trimed_x[-1] >= 15 and trimed_x[-1] < 75:
            step = 5
        elif trimed_x[-1] >= 75 and trimed_x[-1] < 150:
            step = 10
        else:
            step = 20
        xi = list(range(0, trimed_x[-1]+1, step))
        plt.xticks(xi, xi)
        plt.title('Percentage of answers that are shared by multiple\nexercises for complex_factor_proba=' + str(values[i]))
        plt.savefig('graphs/First_degree_equation/Complex_factor_probability/sharing_ex_proba/FDE_complex_factor_proba_' + str(values[i]) + 'sharing_ex_proba.png', bbox_inches='tight')
        plt.clf()

    config = default_config